"""Local Launcher module.

The Local Launcher module provides PyEnSight with the ability to launch an
EnSight :class:`Session<ansys.pyensight.core.Session>` instance using a
local Ansys installation.

Examples:
>>> from ansys.pyensight.core import LocalLauncher
>>> session = LocalLauncher().start()
"""
import glob
import logging
import os.path
import platform
import re
import shutil
import subprocess
import tempfile
import time
from typing import Optional
import uuid
import warnings

import ansys.pyensight.core as pyensight
from ansys.pyensight.core.common import GRPC_WARNING_MESSAGE, find_unused_ports, grpc_version_check
from ansys.pyensight.core.launcher import Launcher
import ansys.pyensight.core.session
import psutil


class LocalLauncher(Launcher):
    """Creates a ``Session`` instance by launching a local copy of EnSight.

    This class allows you to launch locally a copy of EnSight that supports the
    gRPC interface.  It creates and binds a :class:`Session<ansys.pyensight.core.Session>`
    instance to the created gRPC session and returns that instance.

    Parameters
    ----------
    ansys_installation : str, optional
        Path to the local Ansys installation, including the version
        directory. The default is ``None``, in which case common locations
        are scanned to detect the latest local Ansys installation. The
        ``PYENSIGHT_ANSYS_INSTALLATION`` environmental variable is checked first.
    application : str, optional
        App to launch. The default is ``ensight``, but ``envision`` is
        also an option.
    batch : bool, optional
        Whether to run EnSight (or EnVision) in batch mode. The default
        is ``True``, in which case the full GUI is not presented.
    grpc_use_tcp_sockets :
        If using gRPC, and if True, then allow TCP Socket based connections
        instead of only local connections.
    grpc_allow_network_connections :
        If using gRPC and using TCP Socket based connections, listen on all networks.
    grpc_disable_tls :
        If using gRPC and using TCP Socket based connections, disable TLS.
    grpc_uds_pathname :
        If using gRPC and using Unix Domain Socket based connections, explicitly
        set the pathname to the shared UDS file instead of using the default.
    timeout : float, optional
        Number of seconds to try a gRPC connection before giving up.
        This parameter is defined on the parent ``Launcher`` class,
        where the default is ``120``.
    use_egl : bool, optional
        Whether to use EGL hardware for accelerated graphics. The platform
        must be able to support this hardware. This parameter is defined on
        the parent ``Launcher`` class, where the default is ``False``.
    use_sos : int, optional
        Number of EnSight servers to use for SOS (Server of Server) mode.
        This parameter is defined on the parent ``Launcher`` class, where
        the default is ``None``, in which case SOS mode is not used.
    additional_command_line_options: list, optional
        Additional command line options to be used to launch EnSight.

    Examples
    --------
    >>> from ansys.pyensight.core import LocalLauncher
    >>> # Create one EnSight session
    >>> session1 = LocalLauncher(ansys_installation='/ansys_inc/v232').start()
    >>> # Create a second session (a new LocalLauncher instance is required)
    >>> session2 = LocalLauncher(ansys_installation='/ansys_inc/v232').start()

    WARNING:
    Overriding the default values for these options: grpc_use_tcp_sockets, grpc_allow_network_connections,
    and grpc_disable_tls
    can possibly permit control of this computer and any data which resides on it.
    Modification of this configuration is not recommended.  Please see the
    documentation for your installed product for additional information.
    """

    def __init__(
        self,
        ansys_installation: Optional[str] = None,
        application: Optional[str] = "ensight",
        batch: bool = True,
        grpc_use_tcp_sockets: Optional[bool] = False,
        grpc_allow_network_connections: Optional[bool] = False,
        grpc_disable_tls: Optional[bool] = False,
        grpc_uds_pathname: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        # get the user selected installation directory
        self._install_path: str = self.get_cei_install_directory(ansys_installation)
        # Will this be ensight or envision
        self._application = application
        # EnSight session secret key
        self._secret_key: str = ""
        # temporary directory served by websocketserver
        self._session_directory = None
        # launched process ids
        self._ensight_pid = None
        self._websocketserver_pid = None
        self._webui_pid = None
        # and ports
        self._ports = None
        # Are we running the instance in batch
        self._batch = batch
        self._grpc_use_tcp_sockets = grpc_use_tcp_sockets
        self._grpc_allow_network_connections = grpc_allow_network_connections
        self._grpc_disable_tls = grpc_disable_tls
        self._grpc_uds_pathname = grpc_uds_pathname

    @property
    def application(self):
        """Type of app to launch. Options are ``ensight`` and ``envision``."""
        return self._application

    def launch_webui(self, version, popen_common):
        if os.environ.get("PYENSIGHT_FLUIDSONE_PATH"):
            fluids_one_path = os.environ["PYENSIGHT_FLUIDSONE_PATH"]
        else:
            awp_path = os.path.dirname(self._install_path)
            platf = "winx64" if self._is_windows() else "linx64"
            fluids_one_path = os.path.join(awp_path, "FluidsOne", "server", platf, "fluids_one")
            if self._is_windows():
                fluids_one_path += ".exe"
        cmd = [fluids_one_path, "--main-run-mode", "post"]
        path_to_webui = self._install_path
        # Dev environment
        path_to_webui_internal = os.path.join(
            path_to_webui, f"nexus{version}", f"ansys{version}", "ensight", "WebUI", "web", "ui"
        )
        # Ansys environment
        path_to_webui_ansys = os.path.join(os.path.dirname(path_to_webui), "FluidsOne", "web", "ui")
        path_to_webui = path_to_webui_internal
        if os.path.exists(path_to_webui_ansys):
            path_to_webui = path_to_webui_ansys
        cmd += ["--server-listen-port", str(self._ports[5])]
        cmd += ["--server-web-roots", path_to_webui]
        cmd += ["--ensight-grpc-port", str(self._ports[0])]
        cmd += ["--ensight-html-port", str(self._ports[2])]
        cmd += ["--ensight-ws-port", str(self._ports[3])]
        cmd += ["--ensight-session-directory", self._session_directory]
        cmd += ["--ensight-secret-key", self._secret_key]
        cmd += ["--main-show-gui", "'False'"]
        if "PYENSIGHT_DEBUG" in os.environ:
            try:
                if int(os.environ["PYENSIGHT_DEBUG"]) > 0:
                    del popen_common["stdout"]
                    del popen_common["stderr"]
            except (ValueError, KeyError):
                pass
        popen_common["env"].update(
            {
                "SIMBA_WEBSERVER_TOKEN": self._secret_key,
                "FLUENT_WEBSERVER_TOKEN": self._secret_key,
            }
        )
        self._webui_pid = subprocess.Popen(cmd, **popen_common).pid

    def _grpc_version_check(self):
        """Check if the gRPC security options apply to the EnSight install."""
        buildinfo = os.path.join(self._install_path, "BUILDINFO.txt")
        if not os.path.exists(buildinfo):
            if not os.path.exists(
                os.path.join(os.path.dirname(self._install_path), "licensingclient")
            ):
                # Dev installation. Assume the gRPC security options are available
                return True
            raise RuntimeError("Couldn't find BUILDINFO file, cannot check installation.")
        with open(buildinfo, "r") as buildinfo_file:
            text = buildinfo_file.read()
        internal_version, ensight_full_version = self._get_versionfrom_buildinfo(text)
        return grpc_version_check(internal_version, ensight_full_version)

    def start(self) -> "pyensight.Session":
        """Start an EnSight session using the local EnSight installation.

        This method launches a copy of EnSight locally that supports the
        gRPC interface. It creates and binds a ``Session`` instance to the
        created gRPC session and returns that session.

        Returns
        -------
        obj
            PyEnSight ``Session`` object instance.

        Raises
        ------
        RuntimeError:
            If the necessary number of ports could not be allocated.
        """
        self._has_grpc_changes = self._grpc_version_check()
        if not self._has_grpc_changes:
            warnings.warn(GRPC_WARNING_MESSAGE)
        tmp_session = super().start()
        if tmp_session:
            return tmp_session
        if self._ports is None:
            # session directory and UUID
            self._secret_key = str(uuid.uuid1())
            self.session_directory = tempfile.mkdtemp(prefix="pyensight_")
            if (
                not self._grpc_uds_pathname
                and not self._grpc_use_tcp_sockets
                and not self._is_windows()
            ):
                self._grpc_uds_pathname = os.path.join(self.session_directory, "pyensight")

            # gRPC port, VNC port, websocketserver ws, websocketserver html
            to_avoid = self._find_ports_used_by_other_pyensight_and_ensight()
            num_ports = 5
            if self._launch_webui:
                num_ports = 6
            self._ports = find_unused_ports(num_ports, avoid=to_avoid)
            if self._ports is None:
                raise RuntimeError("Unable to allocate local ports for EnSight session")
            is_windows = self._is_windows()

            # Launch EnSight
            # create the environmental variables
            local_env = os.environ.copy()
            if not local_env.get("ENSIGHT_GRPC_DISABLE_SECURITY_TOKEN"):
                local_env["ENSIGHT_SECURITY_TOKEN"] = self._secret_key
            local_env["WEBSOCKETSERVER_SECURITY_TOKEN"] = self._secret_key
            local_env["ENSIGHT_SESSION_TEMPDIR"] = self.session_directory
            # If for some reason, the ENSIGHT_ANSYS_LAUNCH is set previously,
            # honor that value, otherwise set it to "pyensight".  This allows
            # for an environmental setup to set the value to something else
            # (e.g. their "app").
            if "ENSIGHT_ANSYS_LAUNCH" not in local_env:
                local_env["ENSIGHT_ANSYS_LAUNCH"] = "pyensight"

            # build the EnSight command
            exe = os.path.join(self._install_path, "bin", self.application)
            cmd = [exe]
            if self._batch:
                cmd.append("-batch")
            else:
                cmd.append("-no_start_screen")
            cmd.extend(["-grpc_server", str(self._ports[0])])
            if self._has_grpc_changes:
                if self._grpc_use_tcp_sockets:
                    cmd.append("-grpc_use_tcp_sockets")
                if self._grpc_allow_network_connections:
                    cmd.append("-grpc_allow_network_connections")
                if self._grpc_disable_tls:
                    cmd.append("-grpc_disable_tls")
                if self._grpc_uds_pathname:
                    cmd.append("-grpc_uds_pathname")
                    cmd.append(self._grpc_uds_pathname)
            vnc_url = f"vnc://%%3Frfb_port={self._ports[1]}%%26use_auth=0"
            cmd.extend(["-vnc", vnc_url])
            cmd.extend(["-ports", str(self._ports[4])])
            if self._additional_command_line_options:
                cmd.extend(self._additional_command_line_options)

            use_egl = self._use_egl()

            # to aid in debugging, PYENSIGHT_DEBUG can be set to a non-zero integer
            popen_common = dict(
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                cwd=self.session_directory,
                env=local_env,
            )
            if "PYENSIGHT_DEBUG" in os.environ:
                try:
                    if int(os.environ["PYENSIGHT_DEBUG"]) > 0:
                        del popen_common["stdout"]
                        del popen_common["stderr"]
                except ValueError:
                    pass

            if is_windows:
                cmd[0] += ".bat"
            if use_egl:
                cmd.append("-egl")
            if self._use_sos:
                cmd.append("-sos")
                if not self._use_mpi:
                    cmd.append("-nservers")
                    cmd.append(str(int(self._use_sos)))
                else:
                    cmd.append(f"--np={int(self._use_sos)+1}")
                    cmd.append(f"--mpi={self._use_mpi}")
                    cmd.append(f"--ic={self._interconnect}")
                    hosts = ",".join(self._server_hosts)
                    cmd.append(f"--cnf={hosts}")
            if self._liben_rest:
                cmd.extend(["-rest_server", str(self._ports[2])])

            # cmd.append("-minimize_console")
            logging.debug(f"Starting EnSight with : {cmd}\n")
            self._ensight_pid = subprocess.Popen(cmd, **popen_common).pid

            # Launch websocketserver

            # find websocketserver script
            found_scripts = glob.glob(
                os.path.join(self._install_path, "nexus*", "nexus_launcher", "websocketserver.py")
            )
            if not found_scripts:
                raise RuntimeError("Unable to find websocketserver script")
            # If more than one nexus directory is found, find the one that corresponds
            # to the version that should be used. Otherwise, just take the first one found.
            # This is likely to only happen for developer installations or build areas.
            idx = 0
            try:
                found_scripts_len = len(found_scripts)
                if found_scripts_len > 1:
                    version_str = str(pyensight.__ansys_version__)
                    for i in range(found_scripts_len):
                        if version_str in found_scripts[i]:
                            idx = i
                            break
            except Exception:
                pass
            websocket_script = found_scripts[idx]
            version = re.findall(r"nexus(\d+)", websocket_script)[0]
            # build the commandline
            if not self._liben_rest:
                cmd = [os.path.join(self._install_path, "bin", "cpython"), websocket_script]
                if is_windows:
                    cmd[0] += ".bat"
                cmd.extend(["--http_directory", self.session_directory])
                # http port
                cmd.extend(["--http_port", str(self._ports[2])])
                # vnc port
                cmd.extend(["--client_port", str(self._ports[1])])
                if self._enable_rest_api:
                    # grpc port
                    cmd.extend(["--grpc_port", str(self._ports[0])])
                    if self._has_grpc_changes:
                        if self._grpc_use_tcp_sockets:
                            cmd.append("--grpc_use_tcp_sockets")
                        if self._grpc_allow_network_connections:
                            cmd.append("--grpc_allow_network_connections")
                        if self._grpc_disable_tls:
                            cmd.append("--grpc_disable_tls")
                        if self._grpc_uds_pathname:
                            cmd.append("--grpc_uds_pathname")
                            cmd.append(self._grpc_uds_pathname)
                # EnVision sessions
                cmd.extend(["--local_session", "envision", "5"])
                if int(version) > 252 and self._rest_ws_separate_loops:
                    cmd.append("--separate_loops")
                cmd.extend(["--security_token", self._secret_key])
                # websocket port
                if int(version) > 252 and self._do_not_start_ws:
                    cmd.append("-1")
                else:
                    cmd.append(str(self._ports[3]))
                logging.debug(f"Starting WSS: {cmd}\n")
                if is_windows:
                    startupinfo = subprocess.STARTUPINFO()
                    startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                    popen_common["startupinfo"] = startupinfo
                self._websocketserver_pid = subprocess.Popen(cmd, **popen_common).pid

        # build the session instance
        logging.debug(
            f"Creating session with ports for grpc:{self._ports[0]}\n"
            + f"html:{self._ports[2]} ws:{self._ports[3]}\n"
            + f"key:{self._secret_key}\n"
        )
        use_sos = False
        if self._use_sos:
            use_sos = True

        # need to use Session like this for mock testing this class
        session = ansys.pyensight.core.session.Session(
            host="127.0.0.1",
            grpc_port=self._ports[0],
            grpc_use_tcp_sockets=self._grpc_use_tcp_sockets,
            grpc_allow_network_connections=self._grpc_allow_network_connections,
            grpc_disable_tls=self._grpc_disable_tls,
            grpc_uds_pathname=self._grpc_uds_pathname,
            html_port=self._ports[2],
            ws_port=self._ports[3],
            install_path=self._install_path,
            secret_key=self._secret_key,
            timeout=self._timeout,
            sos=use_sos,
            rest_api=self._enable_rest_api,
            webui_port=self._ports[5] if self._launch_webui else None,
        )
        session.launcher = self
        self._sessions.append(session)
        if self._launch_webui:
            self.launch_webui(version, popen_common)
        return session

    @staticmethod
    def _kill_process_unix(pid):
        external_kill = ["kill", "-9", str(pid)]
        process = psutil.Popen(external_kill, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
        process.wait()

    @staticmethod
    def _kill_process_windows(pid):
        external_kill = ["taskkill", "/F", "/PID", str(pid)]
        process = psutil.Popen(external_kill, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
        process.wait()

    def _kill_process_by_pid(self, pid):
        if self._is_windows():
            self._kill_process_windows(pid)
        else:
            self._kill_process_unix(pid)

    def _kill_process_tree(self, pid):
        try:
            parent = psutil.Process(pid)
            for child in parent.children(recursive=True):
                try:
                    self._kill_process_by_pid(child.pid)
                    child.kill()
                except (psutil.AccessDenied, psutil.ZombieProcess, OSError, psutil.NoSuchProcess):
                    continue
            self._kill_process_by_pid(parent.pid)
            parent.kill()
        except (psutil.AccessDenied, psutil.ZombieProcess, OSError, psutil.NoSuchProcess):
            pass

    def stop(self) -> None:
        """Release any additional resources allocated during launching."""
        maximum_wait_secs = 120.0
        start_time = time.time()
        while (time.time() - start_time) < maximum_wait_secs:
            try:
                shutil.rmtree(self.session_directory)
                self._ports = None
                super().stop()
                return
            except PermissionError:
                pass
            except FileNotFoundError:
                pass
            except Exception:
                raise
        raise RuntimeError(f"Unable to remove {self.session_directory} in {maximum_wait_secs}s")

    def close(self, session):
        """Shut down the launched EnSight session.

        This method closes all associated sessions and then stops the
        launched EnSight instance.

        Parameters
        ----------
        session : ``pyensight.Session``
            Session to close.

        Raises
        ------
        RuntimeError
            If the session was not launched by this launcher.

        """
        if self._websocketserver_pid:
            self._kill_process_tree(self._websocketserver_pid)
        return super().close(session)

    @staticmethod
    def get_cei_install_directory(ansys_installation: Optional[str]) -> str:
        """Get the Ansys distribution CEI directory to use.

        Parameters
        ----------
        ansys_installation : str, optional
            Path to the local Ansys installation, including the version
            directory. The default is ``None``, in which case common locations
            are scanned to detect the latest local Ansys installation. The
            ``PYENSIGHT_ANSYS_INSTALLATION`` environmental variable is checked first.

        Returns
        -------
        str
            Validated installation directory, which contains ``bin/ensight``.

        Raises
        ------
        RuntimeError:
            If the installation directory does not point to a
            valid EnSight installation.
        """
        dirs_to_check = []
        if ansys_installation:
            # User passed directory
            dirs_to_check.append(os.path.join(ansys_installation, "CEI"))
            dirs_to_check.append(ansys_installation)
        else:
            # Environmental variable
            if "PYENSIGHT_ANSYS_INSTALLATION" in os.environ:
                env_inst = os.environ["PYENSIGHT_ANSYS_INSTALLATION"]
                dirs_to_check.append(env_inst)
                # Note: PYENSIGHT_ANSYS_INSTALLATION is designed for devel builds
                # where there is no CEI directory, but for folks using it in other
                # ways, we'll add that one too, just in case.
                dirs_to_check.append(os.path.join(env_inst, "CEI"))
            # 'enve' home directory (running in local distro)
            try:
                import enve

                dirs_to_check.append(enve.home())
            except ModuleNotFoundError:
                pass
            # Look for Ansys install using target version number
            version = pyensight.__ansys_version__
            if f"AWP_ROOT{version}" in os.environ:
                dirs_to_check.append(os.path.join(os.environ[f"AWP_ROOT{version}"], "CEI"))
            # Common, default install locations
            install_dir = f"/ansys_inc/v{version}/CEI"
            if platform.system().startswith("Wind"):
                install_dir = rf"C:\Program Files\ANSYS Inc\v{version}\CEI"
            dirs_to_check.append(install_dir)

        for install_dir in dirs_to_check:
            launch_file = os.path.join(install_dir, "bin", "ensight")
            if os.path.exists(launch_file):
                return install_dir

        raise RuntimeError(f"Unable to detect an EnSight installation in: {dirs_to_check}")

    def _is_system_egl_capable(self) -> bool:
        """Check if the system supports the EGL launch.

        Returns
        -------
        bool
            ``True`` if the system supports the EGL launch, ``False`` otherwise.
        """
        if self._is_windows():
            return False
        egl_test_path = os.path.join(self._install_path, "bin", "cei_egltest")
        egl_proc = subprocess.Popen([egl_test_path], stdout=subprocess.PIPE)
        _, _ = egl_proc.communicate()
        if egl_proc.returncode == 0:
            return True
        return False



from types import ModuleType
from typing import Optional, Union

try:
    import ensight
except ImportError:
    from ansys.api.pyensight import ensight_api


class Adr:
    """Provides the ``ensight.utils.adr`` interface.

    The methods in this class implement simplified interfaces connect EnSight
    to an existing ADR report.
    With connect_to_existing_adr_report() the connection is done, while with
    generate_adr_report() the report is generated.

    Parameters
    ----------
    interface :
        Entity that provides the ``ensight`` namespace. In the case of
        EnSight Python, the ``ensight`` module is passed. In the case
        of PyEnSight, ``Session.ensight`` is passed.
    """

    def __init__(self, interface: Union["ensight_api.ensight", "ensight"]):
        self._ensight = interface
        self._adr_report_connected = False
        self._url: Optional[str] = None

    def connect_to_existing_adr_report(
        self, url: str, username: str = "nexus", password: str = "cei"
    ) -> None:
        """Connect to an existing ADR report

        Parameters
        ----------
        url: string
            the url of the ADR report
        username: string
            the username to connect to the ADR report
        password: string
            the password to connect to the ADR report
        """
        ret = None
        if self._adr_report_connected is True:
            raise RuntimeError(
                f"The PyEnSight session is already connected to the ADR server {self._url}."
            )
        if isinstance(self._ensight, ModuleType):
            self._ensight.core.nexus.ReportServer.get_server().set_URL(url)
            self._ensight.core.nexus.ReportServer.get_server().set_username(username)
            self._ensight.core.nexus.ReportServer.get_server().set_password(password)
            ret = self._ensight.core.nexus.ReportServer.get_server().validate() == 1.0
        else:
            self._ensight._session.cmd(
                f"ensight.core.nexus.ReportServer.get_server().set_URL('{url}')"
            )
            self._ensight._session.cmd(
                f"ensight.core.nexus.ReportServer.get_server().set_username('{username}')"
            )
            self._ensight._session.cmd(
                f"ensight.core.nexus.ReportServer.get_server().set_password('{password}')"
            )
            ret = self._ensight._session.cmd(
                "ensight.core.nexus.ReportServer.get_server().validate() == 1.0"
            )
        if ret is True:
            self._adr_report_connected = True
            self._url = url
        else:
            raise RuntimeError("Couldn't connect to the ADR report server.")

    def generate_adr_report(self):
        """Generate the ADR report with the current states."""
        if self._adr_report_connected is False:
            raise RuntimeError(
                "No ADR report is connected to the PyEnSight session. Please generated one and connect with connect_to_existing_adr_report()"
            )
        if len(self._ensight.objs.core.STATES) > 0:
            self._ensight.objs.core.STATES[0].generate_report()
        else:
            raise RuntimeError("No states are available to generate the report with.")

    def is_connected_to_adr_report(self):
        """True if the PyEnSight session is already connected to an ADR report server."""
        if self._adr_report_connected is True:
            print(f"The PyEnSight session is connected to the ADR report server {self._url}")
        return self._adr_report_connected



"""
The ``libuserd`` module allows PyEnSight to directly access EnSight
user-defined readers (USERD).  Any file format for which EnSight
uses a USERD interface can be read using this API

Examples
--------

>>> from ansys.pyensight.core import libuserd
>>> userd = libuserd.LibUserd()
>>> userd.initialize()
>>> print(userd.library_version())
>>> datafile = "/example/data/CFX/Axial_001.res"
>>> readers = userd.query_format(datafile)
>>> data = readers[0].read_dataset(datafile)
>>> print(data.parts())
>>> print(data.variables())
>>> userd.shutdown()

"""
import enum
import logging
import os
import platform
import shutil
import subprocess
import tempfile
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
import uuid
import warnings

from ansys.api.pyensight.v0 import libuserd_pb2, libuserd_pb2_grpc
from ansys.pyensight.core.common import (
    find_unused_ports,
    get_file_service,
    launch_enshell_interface,
    populate_service_host_port,
    pull_image,
)
import grpc
import numpy
import psutil
import requests

try:
    import docker
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError("The docker module must be installed for DockerLauncher")
except Exception:  # pragma: no cover
    raise RuntimeError("Cannot initialize Docker")


if TYPE_CHECKING:
    from docker import DockerClient
    from docker.models.containers import Container
    from enshell_grpc import EnShellGRPC

# This code is currently in development/beta state
warnings.warn(
    "The libuserd interface/API is still under active development and should be considered beta.",
    stacklevel=2,
)


def _build_enum(name: str, pb_enum: Any, flag: bool = False) -> Union[enum.IntEnum, enum.IntFlag]:
    values = {}
    for v in pb_enum:
        values[v[0]] = v[1]
    if flag:
        return enum.IntFlag(name, values)
    return enum.IntEnum(name, values)


ErrorCodes = _build_enum("ErrorCodes", libuserd_pb2.ErrorCodes.items())
ElementType = _build_enum("ElementType", libuserd_pb2.ElementType.items())
VariableLocation = _build_enum("VariableLocation", libuserd_pb2.VariableLocation.items())
VariableType = _build_enum("VariableType", libuserd_pb2.VariableType.items())
PartHints = _build_enum("PartHints", libuserd_pb2.PartHints.items(), flag=True)
UpdateHints = _build_enum("UpdateHints", libuserd_pb2.UpdateHints.items(), flag=True)
RankValues = _build_enum("RankValues", libuserd_pb2.RankValues.items())


class LibUserdError(Exception):
    """
    This class is an exception object raised from the libuserd
    library itself (not the gRPC remote interface).  The associated
    numeric LibUserd.ErrorCode is available via the 'code' attribute.

    Parameters
    ----------
    msg : str
        The message text to be included in the exception.

    Attributes
    ----------
    code : int
        The LibUserd ErrorCodes enum value for this error.
    """

    def __init__(self, msg) -> None:
        super(LibUserdError, self).__init__(msg)
        self._code = libuserd_pb2.ErrorCodes.UNKNOWN
        if msg.startswith("LibUserd("):
            try:
                self._code = int(msg[len("LibUserd(") :].split(")")[0])
            except Exception:
                pass

    @property
    def code(self) -> int:
        """The numeric error code: LibUserd.ErrorCodes"""
        return self._code


class Query(object):
    """
    The class represents a reader "query" instance.  It includes
    the query name as well as the preferred titles.  The ``data``
    method may be used to access the X,Y plot values.

    Parameters
    ----------
    userd
        The LibUserd instance this query is associated with.
    pb
        The protobuffer that represents this object.

    Attributes
    ----------
    id : int
        The id of this query.
    name : str
        The name of this query.
    x_title : str
        String to use as the x-axis title.
    y_title : str
        String to use as the y-axis title.
    metadata : Dict[str, str]
        The metadata for this query.
    """

    def __init__(self, userd: "LibUserd", pb: libuserd_pb2.QueryInfo) -> None:
        self._userd = userd
        self.id = pb.id
        self.name = pb.name
        self.x_title = pb.x_title
        self.y_title = pb.y_title
        self.metadata = {}
        for key in pb.metadata.keys():
            self.metadata[key] = pb.metadata[key]

    def __str__(self) -> str:
        return f"Query id: {self.id}, name: '{self.name}'"

    def __repr__(self):
        return f"<{self.__class__.__name__} object, id: {self.id}, name: '{self.name}'>"

    def data(self) -> List["numpy.array"]:
        """
        Get the X,Y values for this query.

        Returns
        -------
        List[numpy.array]
            A list of two numpy arrays [X, Y].
        """
        self._userd.connect_check()
        pb = libuserd_pb2.Query_dataRequest()
        pb.query_id = self.id
        try:
            reply = self._userd.stub.Query_data(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return [numpy.array(reply.x), numpy.array(reply.y)]


class Variable(object):
    """
    The class represents a reader "variable" instance.  It includes
    information about the variable, including it type (vector, scalar, etc)
    location (nodal, elemental, etc), name and units.

    Parameters
    ----------
    userd
        The LibUserd instance this query is associated with.
    pb
        The protobuffer that represents this object.

    Attributes
    ----------
    id : int
        The id of this variable.
    name : str
        The name of this variable.
    unit_label : str
        The unit label of this variable, "Pa" for example.
    unit_dims : str
        The dimensions of this variable, "L/S" for distance per second.
    location : "VariableLocation"
        The location of this variable.
    type : "VariableType"
        The type of this variable.
    time_varying : bool
        True if the variable is time-varying.
    complex : bool
        True if the variable is complex.
    number_of_components : int
        The number of components of this variable.  A scalar is 1 and
        a vector is 3.
    metadata : Dict[str, str]
        The metadata for this query.
    """

    def __init__(self, userd: "LibUserd", pb: libuserd_pb2.VariableInfo) -> None:
        self._userd = userd
        self.id = pb.id
        self.name = pb.name
        self.unit_label = pb.unit_label
        self.unit_dims = pb.unit_dims
        self.location = VariableLocation(pb.location)  # type: ignore
        self.type = VariableType(pb.type)  # type: ignore
        self.time_varying = pb.time_varying
        self.complex = pb.complex
        self.interleave_flag = pb.interleave_flag
        self.number_of_components = pb.number_of_components
        self.metadata = {}
        for key in pb.metadata.keys():
            self.metadata[key] = pb.metadata[key]

    def __str__(self) -> str:
        return f"Variable id: {self.id}, name: '{self.name}', type: {self.type.name}, location: {self.location.name}"

    def __repr__(self):
        return f"<{self.__class__.__name__} object, id: {self.id}, name: '{self.name}'>"


class Part(object):
    """
    This class represents the EnSight notion of a part.  A part is a single mesh consisting
    of a nodal array along with a collection of element specifications.  Methods are provided
    to access the nodes and connectivity as well as any variables that might be defined
    on the nodes or elements of this mesh.

    Parameters
    ----------
    userd
        The LibUserd instance this query is associated with.
    pb
        The protobuffer that represents this object.

    Attributes
    ----------
    id : int
        The id of this part.
    name : str
        The name of this part.
    reader_id : int
        The id of the Reader this part is associated with.
    hints : int
        See: `PartHints`.
    metadata : Dict[str, str]
        The metadata for this query.
    """

    def __init__(self, userd: "LibUserd", pb: libuserd_pb2.PartInfo):
        self._userd = userd
        self.index = pb.index
        self.id = pb.id
        self.name = pb.name
        self.reader_id = pb.reader_id
        self.hints = pb.hints
        self.metadata = {}
        for key in pb.metadata.keys():
            self.metadata[key] = pb.metadata[key]

    def __str__(self):
        return f"Part id: {self.id}, name: '{self.name}'"

    def __repr__(self):
        return f"<{self.__class__.__name__} object, id: {self.id}, name: '{self.name}'>"

    def nodes(self, rank: Optional[int] = None) -> "numpy.array":
        """
        Return the vertex array for the part.

        Parameters
        ----------
        rank : int, optional
            For a dataset using multiple ranks, the rank to return data from.  The
            default is RankValues.SINGLE_RANK.

        Returns
        -------
        numpy.array
            A numpy array of packed values: x,y,z,z,y,z, ...

        Examples
        --------

        >>> part = reader.parts()[0]
        >>> nodes = part.nodes()
        >>> nodes.shape = (len(nodes)//3, 3)
        >>> print(nodes)

        """
        self._userd.connect_check()
        rank = self._userd.rank_check(rank)
        pb = libuserd_pb2.Part_nodesRequest()
        pb.part_id = self.id
        pb.rank = rank
        try:
            stream = self._userd.stub.Part_nodes(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        nodes = numpy.empty(0, dtype=numpy.float32)
        for chunk in stream:
            if len(nodes) < chunk.total_size:
                nodes = numpy.empty(chunk.total_size, dtype=numpy.float32)
            offset = chunk.offset
            values = numpy.array(chunk.xyz)
            nodes[offset : offset + len(values)] = values
        return nodes

    def num_elements(self, rank: Optional[int] = None) -> dict:
        """
        Get the number of elements of a given type in the current part.

        Parameters
        ----------
        rank : int, optional
            For a dataset using multiple ranks, the rank to return data from.  The
            default is RankValues.SINGLE_RANK.

        Returns
        -------
        dict
            A dictionary with keys being the element type and the values being the number of
            such elements.  Element types with zero elements are not included in the dictionary.

        Examples
        --------

        >>> part = reader.parts()[0]
        >>> elements = part.elements()
        >>> for etype, count in elements.items():
        ...  print(libuserd.ElementType(etype).name, count)

        """
        self._userd.connect_check()
        rank = self._userd.rank_check(rank)
        pb = libuserd_pb2.Part_num_elementsRequest()
        pb.part_id = self.id
        pb.rank = rank
        try:
            reply = self._userd.stub.Part_num_elements(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        elements = {}
        for key in reply.element_count.keys():
            if reply.element_count[key] > 0:
                elements[key] = reply.element_count[key]
        return elements

    def element_conn(self, elem_type: int, rank: Optional[int] = None) -> "numpy.array":
        """
        For "zoo" element types, return the part element connectivity for the specified
        element type.

        Parameters
        ----------
        elem_type : int
            The element type.  All but NFACED and NSIDED element types are allowed.
        rank : int, optional
            For a dataset using multiple ranks, the rank to return data from.  The
            default is RankValues.SINGLE_RANK.

        Returns
        -------
        numpy.array
            A numpy array of the node indices.

        Examples
        --------

        >>> part = reader.parts()[0]
        >>> conn = part.element_conn(libuserd.ElementType.HEX08)
        >>> nodes_per_elem = libuserd_instance.nodes_per_element(libuserd.ElementType.HEX08)
        >>> conn.shape = (len(conn)//nodes_per_elem, nodes_per_elem)
        >>> for element in conn:
        ...  print(element)

        """
        if elem_type >= ElementType.NSIDED:  # type: ignore
            raise RuntimeError(f"Element type {elem_type} is not valid for this call")
        self._userd.connect_check()
        rank = self._userd.rank_check(rank)
        pb = libuserd_pb2.Part_element_connRequest()
        pb.part_id = self.id
        pb.type = elem_type
        pb.rank = rank
        try:
            stream = self._userd.stub.Part_element_conn(pb, metadata=self._userd.metadata())
            conn = numpy.empty(0, dtype=numpy.uint32)
            for chunk in stream:
                if len(conn) < chunk.total_size:
                    conn = numpy.empty(chunk.total_size, dtype=numpy.uint32)
                offset = chunk.offset
                values = numpy.array(chunk.connectivity)
                conn[offset : offset + len(values)] = values
        except grpc.RpcError as e:
            error = self._userd.libuserd_exception(e)
            # if we get an "UNKNOWN" error, then return an empty array
            if isinstance(error, LibUserdError):
                if error.code == ErrorCodes.UNKNOWN:  # type: ignore
                    return numpy.empty(0, dtype=numpy.uint32)
            raise error
        return conn

    def element_conn_nsided(
        self, elem_type: int, rank: Optional[int] = None
    ) -> List["numpy.array"]:
        """
        For an N-Sided element type (regular or ghost), return the connectivity information
        for the elements of that type in this part at this timestep.

        Two arrays are returned in a list:

        - num_nodes_per_element : one number per element that represent the number of nodes in that element
        - nodes : the actual node indices

        Arrays are packed sequentially.  Walking the elements sequentially, if the number of
        nodes for an element is 4, then there are 4 entries added to the nodes array
        for that element.

        Parameters
        ----------
        elem_type: int
            NSIDED or NSIDED_GHOST.
        rank : int, optional
            For a dataset using multiple ranks, the rank to return data from.  The
            default is RankValues.SINGLE_RANK.

        Returns
        -------
        List[numpy.array]
            Two numpy arrays: num_nodes_per_element, nodes
        """
        self._userd.connect_check()
        rank = self._userd.rank_check(rank)
        pb = libuserd_pb2.Part_element_conn_nsidedRequest()
        pb.part_id = self.id
        pb.type = elem_type
        pb.rank = rank
        try:
            stream = self._userd.stub.Part_element_conn_nsided(pb, metadata=self._userd.metadata())
            nodes = numpy.empty(0, dtype=numpy.uint32)
            indices = numpy.empty(0, dtype=numpy.uint32)
            for chunk in stream:
                if len(nodes) < chunk.nodes_total_size:
                    nodes = numpy.empty(chunk.nodes_total_size, dtype=numpy.uint32)
                if len(indices) < chunk.indices_total_size:
                    indices = numpy.empty(chunk.indices_total_size, dtype=numpy.uint32)
                if len(chunk.nodes_per_polygon):
                    offset = chunk.nodes_offset
                    values = numpy.array(chunk.nodes_per_polygon)
                    nodes[offset : offset + len(values)] = values
                if len(chunk.node_indices):
                    offset = chunk.indices_offset
                    values = numpy.array(chunk.node_indices)
                    indices[offset : offset + len(values)] = values
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return [nodes, indices]

    def element_conn_nfaced(
        self, elem_type: int, rank: Optional[int] = None
    ) -> List["numpy.array"]:
        """
        For an N-Faced element type (regular or ghost), return the connectivity information
        for the elements of that type in this part at this timestep.

        Three arrays are returned in a list:

        - num_faces_per_element : one number per element that represent the number of faces in that element
        - num_nodes_per_face : for each face, the number of nodes in the face.
        - face_nodes : the actual node indices

        All arrays are packed sequentially.  Walking the elements sequentially, if the number of
        faces for an element is 4, then there are 4 entries added to the num_nodes_per_face array
        for that element.  Likewise, the nodes for each face are appended in order to the
        face_nodes array.

        Parameters
        ----------
        elem_type: int
            NFACED or NFACED_GHOST.
        rank : int, optional
            For a dataset using multiple ranks, the rank to return data from.  The
            default is RankValues.SINGLE_RANK.

        Returns
        -------
        List[numpy.array]
            Three numpy arrays: num_faces_per_element, num_nodes_per_face, face_nodes
        """
        self._userd.connect_check()
        rank = self._userd.rank_check(rank)
        pb = libuserd_pb2.Part_element_conn_nfacedRequest()
        pb.part_id = self.id
        pb.type = elem_type
        pb.rank = rank
        try:
            stream = self._userd.stub.Part_element_conn_nfaced(pb, metadata=self._userd.metadata())
            face = numpy.empty(0, dtype=numpy.uint32)
            npf = numpy.empty(0, dtype=numpy.uint32)
            nodes = numpy.empty(0, dtype=numpy.uint32)
            for chunk in stream:
                if len(face) < chunk.face_total_size:
                    face = numpy.empty(chunk.face_total_size, dtype=numpy.uint32)
                if len(npf) < chunk.npf_total_size:
                    npf = numpy.empty(chunk.npf_total_size, dtype=numpy.uint32)
                if len(nodes) < chunk.nodes_total_size:
                    nodes = numpy.empty(chunk.nodes_total_size, dtype=numpy.uint32)
                if len(chunk.faces_per_element):
                    offset = chunk.face_offset
                    values = numpy.array(chunk.faces_per_element)
                    face[offset : offset + len(values)] = values
                if len(chunk.nodes_per_face):
                    offset = chunk.npf_offset
                    values = numpy.array(chunk.nodes_per_face)
                    npf[offset : offset + len(values)] = values
                if len(chunk.node_indices):
                    offset = chunk.nodes_offset
                    values = numpy.array(chunk.node_indices)
                    nodes[offset : offset + len(values)] = values
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return [face, npf, nodes]

    def variable_values(
        self,
        variable: "Variable",
        elem_type: int = 0,
        imaginary: bool = False,
        component: int = 0,
        rank: Optional[int] = None,
    ) -> "numpy.array":
        """
        Return a numpy array containing the value(s) of a variable.  If the variable is a
        part variable, a single float value is returned.  If the variable is a nodal variable,
        the resulting numpy array will have the same number of values as there are nodes.
        If the variable is elemental, the `elem_type` selects the block of elements to return
        the variable values for (`elem_type` is ignored for other variable types).

        Parameters
        ----------
        variable : Variable
            The variable to return the values for.
        elem_type : int
            Used only if the variable location is elemental, this keyword selects the element
            type to return the variable values for.
        imaginary : bool
            If the variable is of type complex, setting this to True will select the imaginary
            portion of the data.
        component : int
            Select the channel for a multivalued variable type.  For example, if the variable
            is a vector, setting component to 1 will select the 'Y' component.
        rank : int, optional
            For a dataset using multiple ranks, the rank to return data from.  The
            default is RankValues.SINGLE_RANK.

        Returns
        -------
        numpy.array
            A numpy array or a single scalar float.
        """
        self._userd.connect_check()
        rank = self._userd.rank_check(rank)
        pb = libuserd_pb2.Part_variable_valuesRequest()
        pb.part_id = self.id
        pb.var_id = variable.id
        pb.type = elem_type
        pb.component = component
        pb.complex = imaginary
        pb.rank = rank
        try:
            stream = self._userd.stub.Part_variable_values(pb, metadata=self._userd.metadata())
            v = numpy.empty(0, dtype=numpy.float32)
            for chunk in stream:
                if len(v) < chunk.total_size:
                    v = numpy.empty(chunk.total_size, dtype=numpy.float32)
                offset = chunk.offset
                values = numpy.array(chunk.values)
                v[offset : offset + len(values)] = values
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return v

    def rigid_body_transform(self) -> dict:
        """
        Return the rigid body transform for this part at the current timestep.  The
        returned dictionary includes the following fields:

        - "translation" : Translation 3 floats x,y,z
        - "euler_value" : Euler values 4 floats e0,e1,e2,e3
        - "center_of_gravity" : Center of transform 3 floats x,y,z
        - "rotation_order" : The order rotations are applied 1 float
        - "rotation_angles" : The rotations in radians 3 floats rx,ry,rz

        Returns
        -------
        dict
            The transform dictionary.
        """
        self._userd.connect_check()
        pb = libuserd_pb2.Part_rigid_body_transformRequest()
        pb.part_id = self.id
        try:
            reply = self._userd.stub.Part_rigid_body_transform(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        out = {
            "translation": numpy.array(reply.transform.translation),
            "euler_value": numpy.array(reply.transform.euler_value),
            "center_of_gravity": numpy.array(reply.transform.center_of_gravity),
            "rotation_order": reply.transform.rotation_order,
            "rotation_angles": numpy.array(reply.transform.rotation_angles),
        }
        return out


class Reader(object):
    """
    This class represents is an instance of a user-defined reader that is actively reading a
    dataset.

    Parameters
    ----------
    userd
        The LibUserd instance this query is associated with.
    pb
        The protobuffer that represents this object.

    Attributes
    ----------
    unit_system : str
        The units system provided by the dataset.
    metadata : Dict[str, str]
        The metadata for this query.
    reader_api_version : float
        The reader USERD API version number
    number_of_ranks : int
        The number of ranks being used by the dataset reader

    Notes
    -----
    There can only be one reader active in a single `LibUserd` instance.

    """

    def __init__(self, userd: "LibUserd", pb: libuserd_pb2.Reader) -> None:
        self._userd = userd
        self.unit_system = pb.unit_system
        self.reader_api_version = pb.reader_api_version
        self.number_of_ranks = pb.number_of_ranks
        self.metadata = {}
        for key in pb.metadata.keys():
            self.metadata[key] = pb.metadata[key]
        self.raw_metadata = pb.raw_metadata
        self._timesets: List["numpy.array"] = []
        self._update_timesets()

    def _update_timesets(self) -> None:
        """
        To simplify the interface to time, the timesets are all queried and
        cached.  Additionally, a "common timeset" is generated that combines
        all the timevalues from all timesets into a single timeset which is
        saved as "timeset 0".

        This method reads all the timesets and generates the common timeset.
        """
        if len(self._timesets):
            return
        num_timesets = self.get_number_of_time_sets()
        # The common timeset
        common = set()
        self._timesets = [numpy.array([], dtype="float32")]
        # The other timesets
        for ts in range(1, num_timesets + 1):
            v = self.timevalues(timeset=ts)
            # merge into the common timeset
            common.update(v)
            self._timesets.append(v)
        self._timesets[0] = numpy.array(sorted(list(common)))

    def _common_set_step(self, s: int) -> None:
        """
        When the common timeset is used in a ``set_timestep()`` call, this
        method selects the time value from the common timeset and then calls
        ``_common_set_time()`` to change the current simulation time.

        Parameters
        ----------
        s : int
            The index (timestep) in the common timeset to change the current time to.

        Raises
        ------
        RuntimeError
            If the timestep index is invalid.

        """
        try:
            v = self._timesets[0][s]
            self._common_set_time(v)
        except IndexError:
            raise RuntimeError(f"Invalid step number {s}.") from None

    def _common_set_time(self, t: float) -> None:
        """
        Change the current time value to the passed time value.  This method
        walks all the timesets.  It selects the largest time value in each timeset
        that is less than or equal to the specified time value.  It then sets
        the time value for each timeset accordingly.

        Parameters
        ----------
        t : float
            The time value (in the common timeset) to change the reader simulation time to.

        """
        # given the time float from the common timeline,
        # change the timestep in all the timesets to match.
        for timeset in range(1, len(self._timesets)):
            # check for perfect match first (avoids rounding)
            where = numpy.where(self._timesets[timeset] == t)
            if len(where[0]):
                timestep = where[0][0]
            else:
                timestep = numpy.searchsorted(self._timesets[timeset], t)
                timestep = min(timestep, len(self._timesets[timeset]) - 1)
            self.set_timestep(timestep, timeset=timeset)

    def parts(self) -> List[Part]:
        """
        Get a list of the parts this reader can access.

        Returns
        -------
        List[Part]
            A list of Part objects.
        """
        self._userd.connect_check()
        pb = libuserd_pb2.Reader_partsRequest()
        try:
            parts = self._userd.stub.Reader_parts(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        out = []
        for part in parts.part_list:
            out.append(Part(self._userd, part))
        return out

    def variables(self) -> List[Variable]:
        """
        Get a list of the variables this reader can access.

        Returns
        -------
        List[Variable]
            A list of Variable objects.
        """
        self._userd.connect_check()
        pb = libuserd_pb2.Reader_variablesRequest()
        try:
            variables = self._userd.stub.Reader_variables(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        out = []
        for variable in variables.variable_list:
            out.append(Variable(self._userd, variable))
        return out

    def queries(self) -> List[Query]:
        """
        Get a list of the queries this reader can access.

        Returns
        -------
        List[Query]
            A list of Query objects.
        """
        self._userd.connect_check()
        pb = libuserd_pb2.Reader_queriesRequest()
        try:
            queries = self._userd.stub.Reader_queries(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        out = []
        for query in queries.query_list:
            out.append(Query(self._userd, query))
        return out

    def get_number_of_time_sets(self) -> int:
        """
        Get the number of timesets in the dataset.

        Returns
        -------
        int
            The number of timesets.
        """
        if len(self._timesets):
            return len(self._timesets) - 1
        self._userd.connect_check()
        pb = libuserd_pb2.Reader_get_number_of_time_setsRequest()
        try:
            reply = self._userd.stub.Reader_get_number_of_time_sets(
                pb, metadata=self._userd.metadata()
            )
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return reply.number_of_timesets

    def timevalues(self, timeset: int = 0) -> List[float]:
        """
        Get a list of the time step values in this dataset for the specified timeset.
        The default timeset is ``0`` which is a special, "common" timeset formed by
        merging all the timesets in the data into a single timeset.

        Parameters
        ----------
        timeset : int, optional
            The timestep to query (default is 0)

        Returns
        -------
        numpy.array
            The simulation time value floats.
        """
        if timeset == 0:
            try:
                return self._timesets[timeset]
            except IndexError:
                return numpy.array([], dtype="float32")
        self._userd.connect_check()
        pb = libuserd_pb2.Reader_timevaluesRequest()
        pb.timeset_number = timeset
        try:
            timevalues = self._userd.stub.Reader_timevalues(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return numpy.array(timevalues.time_values)

    def set_timevalue(self, timevalue: float, timeset: int = 0) -> None:
        """
        Change the current time within the selected timeset to the specified value.
        The default timeset selected is the merged "common" timeset ``0``.  If the
        "common" timeset is used, the appropriate time value will be set for
        all timesets by this method.

        Parameters
        ----------
        timevalue : float
            The time value to change the timestep closest to.
        timeset : int, optional
            The timeset to change (default is 0)

        Examples
        --------
        >>> from ansys.pyensight.core import libuserd
        >>> import numpy
        >>> s = libuserd.LibUserd()
        >>> s.initialize()
        >>> opt = {'Long names': 0, 'Number of timesteps': 5, 'Number of scalars': 3,
        ...        'Number of spheres': 2, 'Number of cubes': 2}
        >>> d = s.load_data("foo", file_format="Synthetic", reader_options=opt)
        >>> parts = d.parts()
        >>> for t in d.timevalues():
        ...    d.set_timevalue(t)
        ...    for p in parts:
        ...        nodes = p.nodes()
        ...        nodes.shape = (len(nodes)//3, 3)
        ...        centroid = numpy.average(nodes, 0)
        ...        print(f"Time: {t} Part: {p.name} Centroid: {centroid}")
        >>> s.shutdown()

        """
        if timeset == 0:
            self._common_set_time(timevalue)
            return
        self._userd.connect_check()
        pb = libuserd_pb2.Reader_set_timevalueRequest()
        pb.timeset_number = timeset
        pb.time_value = timevalue
        try:
            _ = self._userd.stub.Reader_set_timevalue(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)

    def set_timestep(self, timestep: int, timeset: int = 0) -> None:
        """
        Change the current time to the specified timestep.  This call is the same as:
        ``reader.set_timevalue(reader.timevalues()[timestep])``.
        The default timeset selected is the merged "common" timeset ``0``.  If the
        "common" timeset is used, the appropriate time value will be set for
        all timesets by this method.

        Parameters
        ----------
        timestep : int
            The timestep to change to.
        timeset : int, optional
            The timeset to change (default is 0)
        """
        if timeset == 0:
            self._common_set_step(timestep)
            return
        self._userd.connect_check()
        pb = libuserd_pb2.Reader_set_timestepRequest()
        pb.timeset_number = timeset
        pb.time_step = timestep
        try:
            _ = self._userd.stub.Reader_set_timestep(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)

    def is_geometry_changing(self) -> bool:
        """
        Check to see if the geometry in this dataset is changing. over time

        Returns
        -------
        bool
            True if the geometry is changing, False otherwise.
        """
        self._userd.connect_check()
        pb = libuserd_pb2.Reader_is_geometry_changingRequest()
        try:
            reply = self._userd.stub.Reader_is_geometry_changing(
                pb, metadata=self._userd.metadata()
            )
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return reply.is_geometry_changing

    def dynamic_update_check(self, changes_allowed: int) -> int:
        """
        When this method is called, it allows the reader to change aspects of
        dataset.  If the reader changes the time steps, mesh data or the
        dataset structure (variable, part or query lists) this function
        will return the nature of the change.  The returned bits will always be
        a subset of the input allowed bits.

        Parameters
        ----------
        changes_allowed : int
            Bitfield formed by oring UpdateHints values for the things that are allowed to change.

        Returns
        -------
        int
            A bitfield formed by oring UpdateHints values for the things that changed.
        """
        self._userd.connect_check()
        pb = libuserd_pb2.Reader_dynamic_update_checkRequest()
        pb.changes_allowed = changes_allowed
        try:
            reply = self._userd.stub.Reader_dynamic_update_check(
                pb, metadata=self._userd.metadata()
            )
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return reply.changed

    def variable_value(self, variable: "Variable", rank: Optional[int] = None) -> float:
        """
        For any "case" variable (e.g. time), the value of the variable.

        Parameters
        ----------
        variable
            The variable to query.  Note, this variable location must be on a CASE.
        rank : int, optional
            For a dataset using multiple ranks, the rank to return data from.  The
            default is RankValues.SINGLE_RANK.
        Returns
        -------
        float
            The value of the variable.
        """
        self._userd.connect_check()
        rank = self._userd.rank_check(rank)
        pb = libuserd_pb2.Reader_variable_valueRequest()
        pb.variable_id = variable.id
        pb.rank = rank
        try:
            reply = self._userd.stub.Reader_variable_value(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return reply.value


class ReaderInfo(object):
    """
    This class represents an available reader, before it has been instantiated.
    The read_dataset() function actually tries to open a dataset and returns
    a `Reader` instance that is reading the data.

    The class contains a list of options that can control/configure the reader.
    These include "boolean", "option" and "field" options.  These include defaults
    supplied by the reader. To use these, change the value or value_index fields
    to the desired values before calling `read_dataset`.

    Parameters
    ----------
    userd
        The LibUserd instance this query is associated with.
    pb
        The protobuffer that represents this object.

    Attributes
    ----------
    id : int
        The reader id.
    name : str
        The reader name.
    description : str
        A brief description of the reader and in some cases its operation.
    file_label1 : str
        A string appropriate for a "file select" button for the primary filename.
    file_label2 : str
        A string appropriate for a "file select" button for the secondary filename.
    opt_booleans : List[dict]
        The boolean user options.
    opt_options : List[dict]
        The option user options suitable for display via an option menu.
    opt_fields : List[dict]
        The field user options suitable for display via a text field.
    """

    def __init__(self, userd: "LibUserd", pb: libuserd_pb2.ReaderInfo):
        self._userd = userd
        self.id = pb.id
        self.name = pb.name
        self.description = pb.description
        self.file_label1 = pb.file_label1
        self.file_label2 = pb.file_label2
        self.opt_booleans = []
        for b in pb.options.booleans:
            self.opt_booleans.append(dict(name=b.name, value=b.value, default=b.default_value))
        self.opt_options = []
        for o in pb.options.options:
            values = []
            for v in o.values:
                values.append(v)
            self.opt_options.append(
                dict(name=o.name, values=values, value=o.value_index, default=o.default_value_index)
            )
        self.opt_fields = []
        for f in pb.options.fields:
            self.opt_fields.append(dict(name=f.name, value=f.value, default=f.default_value))

    def read_dataset(self, file1: str, file2: str = "") -> "Reader":
        """
        Attempt to read some files on disk using this reader and the specified options.
        If successful, return an actual reader instance.

        Parameters
        ----------
        file1 : str
            The primary filename (e.g. "foo.cas")
        file2 : str
            An optional secondary filename (e.g. "foo.dat")

        Returns
        -------
        Reader
            An instance of the `Reader` class.
        """
        self._userd.connect_check()
        pb = libuserd_pb2.ReaderInfo_read_datasetRequest()
        pb.filename_1 = file1
        if file2:
            pb.filename_2 = file2
        pb.reader_id = self.id
        options = self._get_option_values()
        for b in options["booleans"]:
            pb.option_values_bools.append(b)
        for o in options["options"]:
            pb.option_values_options.append(o)
        for f in options["fields"]:
            pb.option_values_fields.append(f)
        try:
            reader = self._userd.stub.ReaderInfo_read_dataset(pb, metadata=self._userd.metadata())
        except grpc.RpcError as e:
            raise self._userd.libuserd_exception(e)
        return Reader(self._userd, reader.reader)

    def _get_option_values(self) -> dict:
        """Extract the current option values from the options dictionaries"""
        out = dict()
        booleans = []
        for b in self.opt_booleans:
            booleans.append(b["value"])
        out["booleans"] = booleans
        options = []
        for o in self.opt_options:
            options.append(o["value"])
        out["options"] = options
        fields = []
        for f in self.opt_fields:
            fields.append(f["value"])
        out["fields"] = fields
        return out

    def __str__(self) -> str:
        return f"ReaderInfo id: {self.id}, name: {self.name}, description: {self.description}"

    def __repr__(self):
        return f"<{self.__class__.__name__} object, id: {self.id}, name: '{self.name}'>"


class LibUserd(object):
    """
    LibUserd is the primary interface to the USERD library. All interaction starts at this object.

    Parameters
    ----------
    ansys_installation : str
        Optional location to search for an Ansys software installation.

    Examples
    --------

    >>> from ansys.pyensight.core import libuserd
    >>> l = libuserd.LibUserd()
    >>> l.initialize()
    >>> readers = l.query_format("D:\\data\\Axial_001.res")
    >>> data = readers[0].read_dataset("D:\\data\\Axial_001.res")
    >>> part = data.parts[0]
    >>> print(part, part.nodes())
    >>> l.shutdown()

    """

    def __init__(
        self,
        ansys_installation: str = "",
        use_docker: bool = False,
        data_directory: Optional[str] = None,
        docker_image_name: Optional[str] = None,
        use_dev: bool = False,
        product_version: Optional[str] = None,
        channel: Optional[grpc.Channel] = None,
        pim_instance: Optional[Any] = None,
        timeout: float = 120.0,
        pull_image_if_not_available: bool = False,
        number_of_ranks: int = 1,
    ):
        self._server_pathname: Optional[str] = None
        self._host = "127.0.0.1"
        self._security_token = str(uuid.uuid1())
        self._grpc_port = 0
        self._server_process: Optional[subprocess.Popen] = None
        self._channel: Optional[grpc.Channel] = None
        self._stub = None
        self._security_file: Optional[str] = None
        # Docker attributes
        self._pull_image = pull_image_if_not_available
        self._timeout = timeout
        self._product_version = product_version
        self._data_directory = data_directory
        self._image_name = "ghcr.io/ansys-internal/ensight"
        if use_dev:
            self._image_name = "ghcr.io/ansys-internal/ensight_dev"
        if docker_image_name:
            self._image_name = docker_image_name
        self._docker_client: Optional["DockerClient"] = None
        self._container: Optional["Container"] = None
        self._enshell: Optional["EnShellGRPC"] = None
        self._pim_instance = pim_instance
        self._enshell_grpc_channel: Optional[grpc.Channel] = channel
        self._pim_file_service: Optional[Any] = None
        self._service_host_port: Dict[str, Tuple[str, int]] = {}
        self._number_of_ranks = number_of_ranks
        local_launch = True
        if any([use_docker, use_dev, self._pim_instance]):
            local_launch = False
            self._launch_enshell()
        else:
            # find the pathname to the server
            self._server_pathname = self._find_ensight_server_name(
                ansys_installation=ansys_installation
            )
            if self._server_pathname is None:
                raise RuntimeError("Unable to detect an EnSight server installation.")
        # enums
        self._build_enums()
        if local_launch:
            self._local_launch()
        # Build the gRPC connection
        self._connect()

    def _local_launch(self) -> None:
        """Launch the gRPC server from a local installation."""
        # have the server save status so we can read it later
        with tempfile.TemporaryDirectory() as tmpdirname:
            self._security_file = os.path.join(tmpdirname, "security.grpc")

            # Build the command line
            cmd = [str(self.server_pathname)]
            cmd.extend(["-grpc_server", str(self.grpc_port)])
            cmd.extend(["-security_file", self._security_file])
            env_vars = os.environ.copy()
            if self.security_token:
                env_vars["ENSIGHT_SECURITY_TOKEN"] = self.security_token
            env_vars["ENSIGHT_GRPC_SECURITY_FILE"] = self._security_file
            # start the server
            try:
                self._server_process = subprocess.Popen(
                    cmd,
                    close_fds=True,
                    env=env_vars,
                    stderr=subprocess.DEVNULL,
                    stdout=subprocess.DEVNULL,
                )
            except Exception as error:
                raise error

            start_time = time.time()
            while (self._grpc_port == 0) and (time.time() - start_time < 120.0):
                try:
                    # Read the port and security token from the security file
                    with open(self._security_file, "r") as f:
                        for line in f:
                            line = line.strip()
                            if line.startswith("grpc_port:"):
                                self._grpc_port = int(line[len("grpc_port:") :])
                            elif line.startswith("grpc_password:"):
                                self._security_token = line[len("grpc_password:") :]
                except (OSError, IOError):
                    pass

            # Unable to get the grpc port/password
            if self._grpc_port == 0:
                self.shutdown()
                raise RuntimeError(f"Unable to start the gRPC server ({str(self.server_pathname)})")

    def _build_enums(self) -> None:
        # retained for backward compatibility
        self.ErrorCodes = ErrorCodes
        self.ElementType = ElementType
        self.VariableLocation = VariableLocation
        self.VariableType = VariableType
        self.PartHints = PartHints
        self.UpdateHints = UpdateHints
        self.RankValues = RankValues

    def _pull_docker_image(self) -> None:
        """Pull the docker image if not available"""
        pull_image(self._docker_client, self._image_name)

    def _check_if_image_available(self) -> bool:
        """Check if the input docker image is available."""
        if not self._docker_client:
            return False
        filtered_images = self._docker_client.images.list(filters={"reference": self._image_name})
        if len(filtered_images) > 0:
            return True
        return False

    def _launch_enshell(self) -> None:
        """Create an enshell entry point and use it to launch a Container."""
        if self._pim_instance:
            self._service_host_port = populate_service_host_port(self._pim_instance, {})
            self._pim_file_service = get_file_service(self._pim_instance)
            self._grpc_port = int(self._service_host_port["grpc_private"][1])
            self._host = self._service_host_port["grpc_private"][0]
        else:
            if not self._data_directory:
                self._data_directory = tempfile.mkdtemp(prefix="pyensight_")
            available = self._check_if_image_available()
            if not available and self._pull_image and not self._pim_instance:
                self._pull_docker_image()
            ports = find_unused_ports(2, avoid=[1999])
            self._service_host_port = {
                "grpc": ("127.0.0.1", ports[0]),
                "grpc_private": ("127.0.0.1", ports[1]),
            }
            self._grpc_port = ports[1]
        if not self._pim_instance:
            self._launch_container()
        self._enshell = launch_enshell_interface(
            self._enshell_grpc_channel, self._service_host_port["grpc"][1], self._timeout
        )
        self._cei_home = self._enshell.cei_home()
        self._ansys_version = self._enshell.ansys_version()
        # print("CEI_HOME=", self._cei_home)
        # print("Ansys Version=", self._ansys_version)
        grpc_port = self._service_host_port["grpc_private"][1]
        ensight_args = f"-grpc_server {grpc_port}"
        container_env_str = f"ENSIGHT_SECURITY_TOKEN={self._security_token}\n"
        ret = self._enshell.start_ensight_server(ensight_args, container_env_str)
        if ret[0] != 0:  # pragma: no cover
            self._stop_container_and_enshell()  # pragma: no cover
            raise RuntimeError(
                f"Error starting EnSight Server with args: {ensight_args}"
            )  # pragma: no cover

    def _launch_container(self) -> None:
        """Launch a docker container for the input image."""
        self._docker_client = docker.from_env()
        grpc_port = self._service_host_port["grpc"][1]
        private_grpc_port = self._service_host_port["grpc_private"][1]
        ports_to_map = {
            str(self._service_host_port["grpc"][1]) + "/tcp": str(grpc_port),
            str(self._service_host_port["grpc_private"][1]) + "/tcp": str(private_grpc_port),
        }
        enshell_cmd = "-app -v 3 -grpc_server " + str(grpc_port)
        container_env = {
            "ENSIGHT_SECURITY_TOKEN": self.security_token,
        }
        data_volume = {self._data_directory: {"bind": "/data", "mode": "rw"}}

        if not self._docker_client:
            raise RuntimeError("Could not startup docker.")
        self._container = self._docker_client.containers.run(  # pragma: no cover
            self._image_name,
            command=enshell_cmd,
            volumes=data_volume,
            environment=container_env,
            ports=ports_to_map,
            tty=True,
            detach=True,
            auto_remove=True,
            remove=True,
        )

    def _stop_container_and_enshell(self) -> None:
        """Release any additional resources allocated during launching."""
        if self._enshell:
            if self._enshell.is_connected():  # pragma: no cover
                try:
                    logging.debug("Stopping EnShell.\n")
                    self._enshell.stop_server()
                except Exception:  # pragma: no cover
                    pass  # pragma: no cover
                self._enshell = None
        if self._container:
            try:
                logging.debug("Stopping the Docker Container.\n")
                self._container.stop()
            except Exception:
                pass
            try:
                logging.debug("Removing the Docker Container.\n")
                self._container.remove()
            except Exception:
                pass
            self._container = None

        if self._pim_instance is not None:
            logging.debug("Deleting the PIM instance.\n")
            self._pim_instance.delete()
            self._pim_instance = None

    @property
    def stub(self):
        """A libuserd_pb2_grpc.LibUSERDServiceStub instance bound to a gRPC connection channel"""
        return self._stub

    @property
    def server_pathname(self) -> Optional[str]:
        """The pathanme of the detected EnSight server executable used as the gRPC server"""
        return self._server_pathname

    @property
    def security_token(self) -> str:
        """The current gRPC security token"""
        return self._security_token

    @property
    def grpc_port(self) -> int:
        """The current gRPC port"""
        return self._grpc_port

    def __del__(self) -> None:
        self.shutdown()

    @staticmethod
    def _find_ensight_server_name(ansys_installation: str = "") -> Optional[str]:
        """
        Parameters
        ----------
        ansys_installation : str
            Path to the local Ansys installation, including the version
            directory. The default is ``None``, in which case common locations
            are scanned to detect the latest local Ansys installation. The
            ``PYENSIGHT_ANSYS_INSTALLATION`` environmental variable is checked first.

        Returns
        -------
        str
            The first valid ensight_server found or None

        """
        dirs_to_check = []
        if ansys_installation:
            dirs_to_check.append(ansys_installation)

        if "PYENSIGHT_ANSYS_INSTALLATION" in os.environ:
            env_inst = os.environ["PYENSIGHT_ANSYS_INSTALLATION"]
            dirs_to_check.append(env_inst)
            # Note: PYENSIGHT_ANSYS_INSTALLATION is designed for devel builds
            # where there is no CEI directory, but for folks using it in other
            # ways, we'll add that one too, just in case.
            dirs_to_check.append(os.path.join(env_inst, "CEI"))

        try:
            import enve

            dirs_to_check.append(enve.home())
        except ModuleNotFoundError:
            pass

        if "CEI_HOME" in os.environ:
            env_inst = os.environ["CEI_HOME"]
            dirs_to_check.append(env_inst)

        # Look for most recent Ansys install
        awp_roots = []
        for env_name in dict(os.environ).keys():
            if env_name.startswith("AWP_ROOT"):
                try:
                    version = int(env_name[len("AWP_ROOT") :])
                    # this API is new in 2025 R1 distributions
                    if version >= 251:
                        awp_roots.append(env_name)
                except ValueError:
                    pass
        awp_roots.sort(reverse=True)
        for env_name in awp_roots:
            dirs_to_check.append(os.path.join(os.environ[env_name], "CEI"))

        # check all the collected locations in order
        app_name = "ensight_server"
        if platform.system() == "Windows":
            app_name += ".bat"
        for install_dir in dirs_to_check:
            launch_file = os.path.join(install_dir, "bin", app_name)
            if os.path.isfile(launch_file):
                return launch_file
        return None

    def _is_connected(self) -> bool:
        """Check to see if the gRPC connection is live

        Returns
        -------
        bool
            True if the connection is active.
        """
        return self._channel is not None

    def _connect(self) -> None:
        """Establish the gRPC connection to EnSight

        Attempt to connect to an EnSight gRPC server using the host and port
        established by the constructor.  Note on failure, this function just
        returns, but is_connected() will return False.

        Parameters
        ----------
        timeout: float
            how long to wait for the connection to timeout
        """
        if self._is_connected():
            return
        # set up the channel

        self._channel = grpc.insecure_channel(
            "{}:{}".format(self._host, self._grpc_port),
            options=[
                ("grpc.max_receive_message_length", -1),
                ("grpc.max_send_message_length", -1),
                ("grpc.testing.fixed_reconnect_backoff_ms", 1100),
            ],
        )
        try:
            grpc.channel_ready_future(self._channel).result(timeout=self._timeout)
        except grpc.FutureTimeoutError:  # pragma: no cover
            self._channel = None  # pragma: no cover
            return  # pragma: no cover
        # hook up the stub interface
        self._stub = libuserd_pb2_grpc.LibUSERDServiceStub(self._channel)

    def metadata(self) -> List[Tuple[bytes, Union[str, bytes]]]:
        """Compute the gRPC stream metadata

        Compute the list to be passed to the gRPC calls for things like security
        and the session name.

        Returns
        -------
        List[Tuple[bytes, Union[str, bytes]]]
            A list object of the metadata elements needed in a gRPC call to
            satisfy the EnSight server gRPC requirements.
        """
        ret: List[Tuple[bytes, Union[str, bytes]]] = list()
        s: Union[str, bytes]
        if self._security_token:  # pragma: no cover
            s = self._security_token
            if type(s) == str:  # pragma: no cover
                s = s.encode("utf-8")
            ret.append((b"shared_secret", s))
        return ret

    def libuserd_exception(self, e: "grpc.RpcError") -> Exception:
        """
        Given an exception raised as the result of a gRPC call, return either
        the input exception or a LibUserdError exception object to differentiate
        between gRPC issues and libuserd issues.

        Parameters
        ----------
        e
            The exception raised by a gRPC call.

        Returns
        -------
        Exception
            Either the original exception or a LibUserdError exception instance, depending on
            the original exception message details.
        """
        msg = e.details()
        if msg.startswith("LibUserd("):
            return LibUserdError(msg)
        return e

    def _disconnect(self, no_error: bool = False) -> None:
        """Close down the gRPC connection

        Disconnect all connections to the gRPC server.  Send the shutdown request gRPC command
        to the server first.

        Parameters
        ----------
        no_error
            If true, ignore errors resulting from the shutdown operation.
        """
        if not self._is_connected():  # pragma: no cover
            return
        # Note: this is expected to return an error
        try:
            pb = libuserd_pb2.Libuserd_shutdownRequest()
            self._stub.Libuserd_shutdown(pb, metadata=self.metadata())  # type: ignore
            if self._channel:
                self._channel.close()
        except grpc.RpcError as e:
            if not no_error:
                raise self.libuserd_exception(e)
        finally:
            # clean up control objects
            self._stub = None
            self._channel = None

    def connect_check(self) -> None:
        """
        Verify that there is an active gRPC connection established.  If not raise
        a RuntimeError

        Raises
        ------
        RuntimeError
            If there is no active connection.
        """
        if not self._is_connected():
            raise RuntimeError("gRPC connection not established")

    def rank_check(self, rank: Optional[int]) -> int:
        """
        Validate the specified rank number.  If the rank is None, return 0.

        Parameters
        ----------
        rank: Optional[int]
            The rank number to validate.

        Returns
        -------
        int
            The validated rank number.
        """
        if rank is None:
            return 0
        if int(rank) < 0 or int(rank) >= self._number_of_ranks:
            raise RuntimeError(
                f"Invalid rank number specified: {rank}.  Valid range:[0, {self._number_of_ranks-1}]"
            )
        return int(rank)

    """
    gRPC method bindings
    """

    def shutdown(self) -> None:
        """
        Close any active gRPC connection and shut down the EnSight server.
        The object is no longer usable.
        """
        self._disconnect(no_error=True)
        # Just in case, we will try to kill the server directly as well
        if self._server_process:
            if psutil.pid_exists(self._server_process.pid):
                proc = psutil.Process(self._server_process.pid)
                for child in proc.children(recursive=True):
                    if psutil.pid_exists(child.pid):
                        # This can be a race condition, so it is ok if the child is dead already
                        try:
                            child.kill()
                        except psutil.NoSuchProcess:
                            pass
                # Same issue, this process might already be shutting down, so NoSuchProcess is ok.
                try:
                    proc.kill()
                except psutil.NoSuchProcess:
                    pass
        if self._container:
            self._stop_container_and_enshell()
        self._server_process = None

    def ansys_release_string(self) -> str:
        """
        Return the Ansys release for the library.

        Returns
        -------
        str
            Return a string like "2025 R1"
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_ansys_release_stringRequest()
        try:
            ret = self.stub.Libuserd_ansys_release_string(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        return ret.result

    def ansys_release_number(self) -> int:
        """
        Return the Ansys release number of the library.

        Returns
        -------
        int
            A version number like 251 (for "2025 R1")
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_ansys_release_numberRequest()
        try:
            ret = self.stub.Libuserd_ansys_release_number(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        return ret.result

    def library_version(self) -> str:
        """
        The library version number.  This string is the version of the
        library interface itself.  This is not the same as the version
        number of the Ansys release that corresponds to the library.

        This number follows semantic versioning rules: "1.0.0" or
        "0.4.3-rc.1" would be examples of valid library_version() strings.

        Returns
        -------
        str
            The library interface version number string.
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_library_versionRequest()
        try:
            ret = self.stub.Libuserd_library_version(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        return ret.result

    def nodes_per_element(self, element_type: int) -> int:
        """
        For a given element type (e.g. HEX20), return the number of nodes used by the element.
        Note, this is not supported for NSIDED and NFACED element types.

        Parameters
        ----------
        element_type : int
            The element type:  ElementType enum value

        Returns
        -------
        int
            Number of nodes per element or 0 if elem_type is not a valid zoo element type.
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_nodes_per_elementRequest()
        pb.elemType = element_type
        try:
            ret = self.stub.Libuserd_nodes_per_element(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        return ret.result

    def element_is_ghost(self, element_type: int) -> bool:
        """

        For a given element type (e.g. HEX20), determine if the element type should be considered
        a "ghost" element.

        Parameters
        ----------
        element_type : int
            The element type:  ElementType enum value

        Returns
        -------
        bool
            True if the element is a ghost (or an invalid element type).
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_element_is_ghostRequest()
        pb.elemType = element_type
        try:
            ret = self.stub.Libuserd_element_is_ghost(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        return ret.result

    def element_is_zoo(self, element_type: int) -> bool:
        """
        For a given element type (e.g. HEX20), determine if the element type is zoo or not

        Parameters
        ----------
        element_type : int
            The element type:  ElementType enum value

        Returns
        -------
        bool
            True if the element is a zoo element and false if it is NSIDED or NFACED.
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_element_is_zooRequest()
        pb.elemType = element_type
        try:
            ret = self.stub.Libuserd_element_is_zoo(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        return ret.result

    def element_is_nsided(self, element_type: int) -> bool:
        """
        For a given element type, determine if the element type is n-sided or not

        Parameters
        ----------
        element_type : int
            The element type:  ElementType enum value

        Returns
        -------
        bool
            True if the element is a NSIDED or NSIDED_GHOST and False otherwise.
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_element_is_nsidedRequest()
        pb.elemType = element_type
        try:
            ret = self.stub.Libuserd_element_is_nsided(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        return ret.result

    def element_is_nfaced(self, element_type: int) -> bool:
        """
        For a given element type, determine if the element type is n-faced or not

        Parameters
        ----------
        element_type : int
            The element type:  ElementType enum value

        Returns
        -------
        bool
            True if the element is a NFACED or NFACED_GHOST and False otherwise.
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_element_is_nfacedRequest()
        pb.elemType = element_type
        try:
            ret = self.stub.Libuserd_element_is_nfaced(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        return ret.result

    def number_of_simple_element_types(self) -> int:
        """
        There is a consecutive range of element type enums that are supported by the
        Part.element_conn() method.  This function returns the number of those elements
        and may be useful in common element type handling code.

        Note: The value is effectively int(ElementType.NSIDED).

        Returns
        -------
        int
            The number of zoo element types.
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_number_of_simple_element_typesRequest()
        try:
            ret = self.stub.Libuserd_number_of_simple_element_types(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        return ret.result

    def initialize(self, number_of_ranks: int = 0) -> None:
        """
        This call initializes the libuserd system.  It causes the library to scan for available
        readers and set up any required reduction engine bits.  It can only be called once.

        Parameters
        ----------
        number_of_ranks : int, optional
            The degree of I/O parallelism to read data with.  Zero is serial I/O.  Note: this
            option is not currently implemented and 0 should be used.
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_initializeRequest()
        if number_of_ranks:
            self._number_of_ranks = number_of_ranks
            try:
                pb.parallel_node_count = number_of_ranks
            except Exception:
                # This exception is allowed to support older .proto versions
                pass
        try:
            _ = self.stub.Libuserd_initialize(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)

    def get_all_readers(self) -> List["ReaderInfo"]:
        """
        Return a list of the readers that are available.

        Returns
        -------
        List[ReaderInfo]
            List of all ReaderInfo objects.
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_get_all_readersRequest()
        try:
            readers = self.stub.Libuserd_get_all_readers(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        out = []
        for reader in readers.reader_info:
            out.append(ReaderInfo(self, reader))
        return out

    def query_format(self, name1: str, name2: str = "") -> List["ReaderInfo"]:
        """
        For a given dataset (filename(s)), ask the readers if they should be able to read
        that data.

        Parameters
        ----------
        name1 : str
            Primary input filename

        name2 : str
            Optional, secondary input filename

        Returns
        -------
        List[ReaderInfo]
            List of ReaderInfo objects that might be able to read the dataset
        """
        self.connect_check()
        pb = libuserd_pb2.Libuserd_query_formatRequest()
        pb.name1 = name1
        if name2:
            pb.name2 = name2
        try:
            readers = self.stub.Libuserd_query_format(pb, metadata=self.metadata())
        except grpc.RpcError as e:
            raise self.libuserd_exception(e)
        out = []
        for reader in readers.reader_info:
            out.append(ReaderInfo(self, reader))
        return out

    def load_data(
        self,
        data_file: str,
        result_file: str = "",
        file_format: Optional[str] = None,
        reader_options: Dict[str, Any] = {},
    ) -> "Reader":
        """Use the reader to load a dataset and return an instance
        to the resulting ``Reader`` interface.

        Parameters
        ----------
        data_file : str
            Name of the data file to load.
        result_file : str, optional
            Name of the second data file for dual-file datasets.
        file_format : str, optional
            Name of the USERD reader to use. The default is ``None``,
            in which case libuserd selects a reader.
        reader_options : dict, optional
            Dictionary of reader-specific option-value pairs that can be used
            to customize the reader behavior. The default is ``None``.

        Returns
        -------
        Reader
            Resulting Reader object instance.

        Raises
        ------
        RuntimeError
            If libused cannot guess the file format or an error occurs while the
            data is being read.

        Examples
        --------

        >>> from ansys.pyensight.core import libuserd
        >>> userd = libuserd.LibUserd()
        >>> userd.initialize()
        >>> opt = {'Long names': False, 'Number of timesteps': '10', 'Number of scalars': '3'}
        >>> data = userd.load_data("foo", file_format="Synthetic", reader_options=opt
        >>> print(data.parts())
        >>> print(data.variables())
        >>> userd.shutdown()

        """
        the_reader: Optional[ReaderInfo] = None
        if file_format:
            for reader in self.get_all_readers():
                if reader.name == file_format:
                    the_reader = reader
                    break
            if the_reader is None:
                raise RuntimeError(f"The reader '{file_format}' could not be found.")
        else:
            readers = self.query_format(data_file, name2=result_file)
            if len(readers):
                the_reader = readers[0]
            if the_reader is None:
                raise RuntimeError(f"Unable to find a reader for '{data_file}':'{result_file}'.")
        for key, value in reader_options.items():
            for b in the_reader.opt_booleans:
                if key == b["name"]:
                    b["value"] = bool(value)
            for o in the_reader.opt_options:
                if key == o["name"]:
                    o["value"] = int(value)
            for f in the_reader.opt_fields:
                if key == f["name"]:
                    f["value"] = str(value)
        try:
            output = the_reader.read_dataset(data_file, result_file)
        except Exception:
            raise RuntimeError("Unable to open the specified dataset.") from None

        return output

    @staticmethod
    def _download_files(uri: str, pathname: str, folder: bool = False, override_root: bool = False):
        """Download files from the input uri and save them on the input pathname.

        Parameters:
        ----------

        uri : str
            The uri to get files from
        pathname : str
            The location were to save the files. It could be either a file or a folder.
        folder : bool
            True if the uri will server files from a directory. In this case,
            pathname will be used as the directory were to save the files.
        override_root: bool
            True if the root has been overridden. So don't consider the case
            of getting the download URL from the github API
        """
        if not folder:
            if override_root:
                correct_url = uri
            else:
                correct_url = None
                with requests.get(uri) as r:
                    data = r.json()
                    correct_url = data["download_url"]
                if not correct_url:
                    raise RuntimeError(f"Couldn't retrieve download URL from github uri {uri}")
            with requests.get(correct_url, stream=True) as r:
                with open(pathname, "wb") as f:
                    shutil.copyfileobj(r.raw, f)
        else:
            with requests.get(uri) as r:
                data = r.json()
                os.makedirs(pathname, exist_ok=True)
                for item in data:
                    if item["type"] == "file":
                        file_url = item["download_url"]
                        filename = os.path.join(pathname, item["name"])
                        r = requests.get(file_url, stream=True)
                        with open(filename, "wb") as f:
                            f.write(r.content)

    def file_service(self) -> Optional[Any]:
        """Get the PIM file service object if available."""
        return self._pim_file_service

    def download_pyansys_example(
        self,
        filename: str,
        directory: Optional[str] = None,
        root: Optional[str] = None,
        folder: bool = False,
    ) -> str:
        """Download an example dataset from the ansys/example-data repository.
        The dataset is downloaded local to the EnSight server location, so that it can
        be downloaded even if running from a container.

        Parameters
        ----------
        filename: str
            The filename to download
        directory: str
            The directory to download the filename from
        root: str
            If set, the download will happen from another location
        folder: bool
            If set to True, it marks the filename to be a directory rather
            than a single file

        Returns
        -------
        pathname: str
            The download location, local to the EnSight server directory.
            If folder is set to True, the download location will be a folder containing
            all the items available in the repository location under that folder.

        Examples
        --------
        >>> from ansys.pyensight.core import libuserd
        >>> l = libuserd.LibUserd()
        >>> cas_file = l.download_pyansys_example("mixing_elbow.cas.h5","pyfluent/mixing_elbow")
        >>> dat_file = l.download_pyansys_example("mixing_elbow.dat.h5","pyfluent/mixing_elbow")
        """
        base_uri = "https://api.github.com/repos/ansys/example-data/contents"
        override_root = False
        if not folder:
            if root is not None:
                base_uri = root
                override_root = True
        uri = f"{base_uri}/{filename}"
        if directory:
            uri = f"{base_uri}/{directory}/{filename}"
        # Local installs and PIM instances
        download_path = f"{os.getcwd()}/{filename}"
        if self._container and self._data_directory:
            # Docker Image
            download_path = os.path.join(self._data_directory, filename)
        self._download_files(uri, download_path, folder=folder, override_root=override_root)
        pathname = download_path
        if self._container:
            # Convert local path to Docker mounted volume path
            pathname = f"/data/{filename}"
        return pathname



import hashlib
import json
import logging
import os
import queue
import sys
import threading
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from ansys.api.pyensight.v0 import dynamic_scene_graph_pb2
from ansys.pyensight.core import ensight_grpc
import numpy

original_stderr = sys.stderr
original_stdout = sys.stdout
sys.stderr = open(os.devnull, "w")
sys.stdout = open(os.devnull, "w")
try:
    import dsgutils

    dsgutils_loaded = True
except (ModuleNotFoundError, ImportError, AttributeError):
    dsgutils_loaded = False
finally:
    sys.stderr = original_stderr
    sys.stdout = original_stdout

if TYPE_CHECKING:
    from ansys.pyensight.core import Session


class Part(object):
    def __init__(self, session: "DSGSession"):
        """
        This object roughly represents an EnSight "Part".  It contains the connectivity,
        coordinates, normals and texture coordinate information for one DSG entity

        This object stores basic geometry information coming from the DSG protocol.  The
        update_geom() method can parse an "UpdateGeom" protobuffer and merges the results
        into the Part object.

        Parameters
        ----------
        session:
            The DSG connection session object.
        """
        self.session = session
        self.conn_tris = numpy.array([], dtype="int32")
        self.conn_lines = numpy.array([], dtype="int32")
        self.coords = numpy.array([], dtype="float32")
        self.normals = numpy.array([], dtype="float32")
        self.normals_elem = False
        self.tcoords = numpy.array([], dtype="float32")
        self.tcoords_elem = False
        self.node_sizes = numpy.array([], dtype="float32")
        self.cmd: Optional[Any] = None
        self.hash = hashlib.new("sha256")
        self._material: Optional[Any] = None
        self.reset()

    def reset(self, cmd: Any = None) -> None:
        """
        Reset the part object state to prepare the object
        for a new part representation. Numpy arrays are cleared
        and the state reset.

        Parameters
        ----------
        cmd: Any
            The DSG command that triggered this reset.  Most likely
            this is a UPDATE_PART command.

        """
        self.conn_tris = numpy.array([], dtype="int32")
        self.conn_lines = numpy.array([], dtype="int32")
        self.coords = numpy.array([], dtype="float32")
        self.normals = numpy.array([], dtype="float32")
        self.normals_elem = False
        self.tcoords = numpy.array([], dtype="float32")
        self.tcoords_var_id = None
        self.tcoords_elem = False
        self.node_sizes = numpy.array([], dtype="float32")
        self.hash = hashlib.new("sha256")
        if cmd is not None:
            self.hash.update(cmd.hash.encode("utf-8"))
        self.cmd = cmd
        self._material = None

    def _parse_material(self) -> None:
        """
        Parse the JSON string in the part command material string and
        make the content accessible via material_names() and material().
        """
        if self._material is not None:
            return
        try:
            if self.cmd.material_name:  # type: ignore
                self._material = json.loads(self.cmd.material_name)  # type: ignore
                for key, value in self._material.items():
                    value["name"] = key
            else:
                self._material = {}
        except Exception as e:
            self.session.warn(f"Unable to parse JSON material: {str(e)}")
            self._material = {}

    def material_names(self) -> List[str]:
        """
        Return the list of material names included in the part material.

        Returns
        -------
        List[str]
            The list of defined material names.
        """
        self._parse_material()
        if self._material is None:
            return []
        return list(self._material.keys())

    def material(self, name: str = "") -> dict:
        """
        Return the material dictionary for the specified material name.

        Parameters
        ----------
        name: str
            The material name to query.  If no material name is given, the
            first name in the material_names() list is used.

        Returns
        -------
        dict
            The material description dictionary or an empty dictionary.
        """
        self._parse_material()
        if not name:
            names = self.material_names()
            if len(names):
                name = names[0]
        if self._material is None:
            return {}
        return self._material.get(name, {})

    def update_geom(self, cmd: dynamic_scene_graph_pb2.UpdateGeom) -> None:
        """
        Merge an update geometry command into the numpy buffers being cached in this object

        Parameters
        ----------
        cmd:
            This is an array update command.  It could be for coordinates, normals, variables, connectivity, etc.
        """
        if cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.COORDINATES:
            if self.coords.size != cmd.total_array_size:
                self.coords = numpy.resize(self.coords, cmd.total_array_size)
            self.coords[cmd.chunk_offset : cmd.chunk_offset + len(cmd.flt_array)] = cmd.flt_array
        elif cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.TRIANGLES:
            if self.conn_tris.size != cmd.total_array_size:
                self.conn_tris = numpy.resize(self.conn_tris, cmd.total_array_size)
            self.conn_tris[cmd.chunk_offset : cmd.chunk_offset + len(cmd.int_array)] = cmd.int_array
        elif cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.LINES:
            if self.conn_lines.size != cmd.total_array_size:
                self.conn_lines = numpy.resize(self.conn_lines, cmd.total_array_size)
            self.conn_lines[
                cmd.chunk_offset : cmd.chunk_offset + len(cmd.int_array)
            ] = cmd.int_array
        elif (cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.ELEM_NORMALS) or (
            cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.NODE_NORMALS
        ):
            self.normals_elem = cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.ELEM_NORMALS
            if self.normals.size != cmd.total_array_size:
                self.normals = numpy.resize(self.normals, cmd.total_array_size)
            self.normals[cmd.chunk_offset : cmd.chunk_offset + len(cmd.flt_array)] = cmd.flt_array
        elif (cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.ELEM_VARIABLE) or (
            cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.NODE_VARIABLE
        ):
            # Get the variable definition
            if cmd.variable_id in self.session.variables:
                if self.cmd.color_variableid == cmd.variable_id:  # type: ignore
                    # Receive the colorby var values
                    self.tcoords_elem = (
                        cmd.payload_type == dynamic_scene_graph_pb2.UpdateGeom.ELEM_VARIABLE
                    )
                    if self.tcoords.size != cmd.total_array_size:
                        self.tcoords = numpy.resize(self.tcoords, cmd.total_array_size)
                    self.tcoords[
                        cmd.chunk_offset : cmd.chunk_offset + len(cmd.flt_array)
                    ] = cmd.flt_array

                    # Add the variable hash to the Part's hash, to pick up palette changes
                    var_cmd = self.session.variables.get(cmd.variable_id, None)
                    if var_cmd is not None:
                        self.hash.update(var_cmd.hash.encode("utf-8"))

                if self.cmd.node_size_variableid == cmd.variable_id:  # type: ignore
                    # Receive the node size var values
                    if self.node_sizes.size != cmd.total_array_size:
                        self.node_sizes = numpy.resize(self.node_sizes, cmd.total_array_size)
                    self.node_sizes[
                        cmd.chunk_offset : cmd.chunk_offset + len(cmd.flt_array)
                    ] = cmd.flt_array
        # Combine the hashes for the UpdatePart and all UpdateGeom messages
        self.hash.update(cmd.hash.encode("utf-8"))

    def nodal_surface_rep(self):
        """
        This function processes the geometry arrays and converts them into nodal representation.
        It will duplicate triangles as needed (to preserve element normals) and will convert
        variable data into texture coordinates.

        Returns
        -------
        On failure, the method returns None for the first return value.  The returned tuple is:

        (part_command, vertices, connectivity, normals, tex_coords, var_command)

        part_command: UPDATE_PART command object
        vertices: numpy array of the nodal coordinates
        connectivity: numpy array of the triangle indices into the vertices array
        normals: numpy array of per vertex normal values (optional)
        tcoords: numpy array of per vertex texture coordinates (optional)
        var_command: UPDATE_VARIABLE command object for the variable the texture coordinate correspond to, if any
        """
        if self.cmd is None:
            return None, None, None, None, None, None
        if self.conn_tris.size == 0:
            self.session.log(f"Note: part '{self.cmd.name}' contains no triangles.")
            return None, None, None, None, None, None
        verts = self.coords
        _ = self._normalize_verts(verts)

        conn = self.conn_tris
        normals = self.normals
        tcoords = None
        if self.tcoords.size:
            tcoords = self.tcoords
        if self.tcoords_elem or self.normals_elem:
            verts_per_prim = 3
            num_prims = conn.size // verts_per_prim
            # "flatten" the triangles to move values from elements to nodes
            new_verts = numpy.ndarray((num_prims * verts_per_prim * 3,), dtype="float32")
            new_conn = numpy.ndarray((num_prims * verts_per_prim,), dtype="int32")
            new_tcoords = None
            if tcoords is not None:
                # remember that the input values are 1D at this point, we will expand to 2D later
                new_tcoords = numpy.ndarray((num_prims * verts_per_prim,), dtype="float32")
            new_normals = None
            if normals is not None:
                if normals.size == 0:
                    self.session.log("Warning: zero length normals!")
                else:
                    new_normals = numpy.ndarray((num_prims * verts_per_prim * 3,), dtype="float32")
            if dsgutils_loaded:
                dsgutils.build_nodal_surface_rep(
                    verts_per_prim,
                    self.normals_elem,
                    self.tcoords_elem,
                    conn,
                    verts,
                    normals,
                    tcoords,
                    new_conn,
                    new_verts,
                    new_normals,
                    new_tcoords,
                )
            else:
                j = 0
                for i0 in range(num_prims):
                    for i1 in range(verts_per_prim):
                        idx = conn[i0 * verts_per_prim + i1]
                        # new connectivity (identity)
                        new_conn[j] = j
                        # copy the vertex
                        new_verts[j * 3 + 0] = verts[idx * 3 + 0]
                        new_verts[j * 3 + 1] = verts[idx * 3 + 1]
                        new_verts[j * 3 + 2] = verts[idx * 3 + 2]
                        if new_normals is not None:
                            if self.normals_elem:
                                # copy the normal associated with the face
                                new_normals[j * 3 + 0] = normals[i0 * 3 + 0]
                                new_normals[j * 3 + 1] = normals[i0 * 3 + 1]
                                new_normals[j * 3 + 2] = normals[i0 * 3 + 2]
                            else:
                                # copy the same normal as the vertex
                                new_normals[j * 3 + 0] = normals[idx * 3 + 0]
                                new_normals[j * 3 + 1] = normals[idx * 3 + 1]
                                new_normals[j * 3 + 2] = normals[idx * 3 + 2]
                        if new_tcoords is not None:
                            # remember, 1D texture coords at this point
                            if self.tcoords_elem:
                                # copy the texture coord associated with the face
                                new_tcoords[j] = tcoords[i0]
                            else:
                                # copy the same texture coord as the vertex
                                new_tcoords[j] = tcoords[idx]
                        j += 1
            # new arrays.
            verts = new_verts
            conn = new_conn
            normals = new_normals
            if tcoords is not None:
                tcoords = new_tcoords

        var_cmd = None
        # texture coords need transformation from variable value to [ST]
        if tcoords is not None:
            tcoords, var_cmd = self._build_st_coords(tcoords, verts.size // 3)

        self.session.log(
            f"Part '{self.cmd.name}' defined: {self.coords.size // 3} verts, {self.conn_tris.size // 3} tris."
        )
        command = self.cmd

        return command, verts, conn, normals, tcoords, var_cmd

    def _normalize_verts(self, verts: numpy.ndarray) -> float:
        """
        This function scales and translates vertices, so the longest axis in the scene is of
        length 1.0, and data is centered at the origin

        Returns the scale factor
        """
        s = 1.0
        if self.session.normalize_geometry and self.session.scene_bounds is not None:
            num_verts = verts.size // 3
            midx = (self.session.scene_bounds[3] + self.session.scene_bounds[0]) * 0.5
            midy = (self.session.scene_bounds[4] + self.session.scene_bounds[1]) * 0.5
            midz = (self.session.scene_bounds[5] + self.session.scene_bounds[2]) * 0.5
            dx = self.session.scene_bounds[3] - self.session.scene_bounds[0]
            dy = self.session.scene_bounds[4] - self.session.scene_bounds[1]
            dz = self.session.scene_bounds[5] - self.session.scene_bounds[2]
            s = dx
            if dy > s:
                s = dy
            if dz > s:
                s = dz
            if s == 0:
                s = 1.0
            for i in range(num_verts):
                j = i * 3
                verts[j + 0] = (verts[j + 0] - midx) / s
                verts[j + 1] = (verts[j + 1] - midy) / s
                verts[j + 2] = (verts[j + 2] - midz) / s
        return 1.0 / s

    def _build_st_coords(self, tcoords: numpy.ndarray, num_verts: int):
        """
        The Omniverse interface uses 2D texturing (s,t) to reference the texture map.
        This method converts DSG texture coordinates (1D and in "variable" units) into
        2D OpenGL style [0.,1.] normalized coordinate space.  the "t" coordinate will
        always be 0.5.

        Parameters
        ----------
        tcoords: numpy.ndarray
            The DSG 1D texture coordinates, which are actually variable values.

        num_verts: int
            The number of vertices in the mesh.

        Returns
        -------
        numpy.ndarray, Any
            The ST OpenGL GL texture coordinate array and the variable definition DSG command.
        """
        var_dsg_id = self.cmd.color_variableid  # type: ignore
        var_cmd = self.session.variables[var_dsg_id]
        v_min = None
        v_max = None
        for lvl in var_cmd.levels:
            if (v_min is None) or (v_min > lvl.value):
                v_min = lvl.value
            if (v_max is None) or (v_max < lvl.value):
                v_max = lvl.value
        var_minmax: List[float] = [v_min, v_max]  # type: ignore
        # build a power of two x 1 texture
        num_texels = len(var_cmd.texture) // 4
        if dsgutils_loaded:
            tmp = dsgutils.build_st_coords(tcoords, v_min, v_max, num_texels)
        else:
            half_texel = 1 / (num_texels * 2.0)
            tmp = numpy.ndarray((num_verts * 2,), dtype="float32")
            tmp.fill(0.5)  # fill in the T coordinate...
            tex_width = half_texel * 2 * (num_texels - 1)  # center to center of num_texels
            # if the range is 0, adjust the min by -1.   The result is that the texture
            # coords will get mapped to S=1.0 which is what EnSight does in this situation
            if (var_minmax[1] - var_minmax[0]) == 0.0:
                var_minmax[0] = var_minmax[0] - 1.0
            var_width = var_minmax[1] - var_minmax[0]
            for idx in range(num_verts):
                # normalized S coord value (clamp)
                s = (tcoords[idx] - var_minmax[0]) / var_width
                if s < 0.0:
                    s = 0.0
                if s > 1.0:
                    s = 1.0
                # map to the texture range and set the S value
                tmp[idx * 2] = s * tex_width + half_texel
        return tmp, var_cmd

    def line_rep(self):
        """
        This function processes the geometry arrays and returns values to represent line data.
        The vertex array embeds the connectivity, so every two points represent a line segment.
        The tcoords similarly follow the vertex array notion.

        Returns
        -------
        On failure, the method returns None for the first return value.  The returned tuple is:

        (part_command, vertices, connectivity, tex_coords, var_command)

        part_command: UPDATE_PART command object
        vertices: numpy array of per-node coordinates (two per line segment)
        tcoords: numpy array of per vertex texture coordinates (optional)
        var_command: UPDATE_VARIABLE command object for the variable the colors correspond to, if any
        """
        if self.cmd is None:
            return None, None, None, None
        if self.cmd.render != self.cmd.CONNECTIVITY:
            # Early out.  Rendering type for this object is a surface rep, not a point rep
            return None, None, None, None

        num_lines = self.conn_lines.size // 2
        if num_lines == 0:
            return None, None, None, None
        verts = numpy.ndarray((num_lines * 2 * 3,), dtype="float32")
        tcoords = None
        if self.tcoords.size:
            tcoords = numpy.ndarray((num_lines * 2,), dtype="float32")
        # TODO: handle elemental line values (self.tcoords_elem) by converting to nodal...
        # if self.tcoords_elem:
        for i in range(num_lines):
            i0 = self.conn_lines[i * 2]
            i1 = self.conn_lines[i * 2 + 1]
            offset = i * 6
            verts[offset + 0] = self.coords[i0 * 3 + 0]
            verts[offset + 1] = self.coords[i0 * 3 + 1]
            verts[offset + 2] = self.coords[i0 * 3 + 2]
            verts[offset + 3] = self.coords[i1 * 3 + 0]
            verts[offset + 4] = self.coords[i1 * 3 + 1]
            verts[offset + 5] = self.coords[i1 * 3 + 2]
            if tcoords is not None:
                # tcoords are 1D at this point
                offset = i * 2
                tcoords[offset + 0] = self.tcoords[i0]
                tcoords[offset + 1] = self.tcoords[i1]

        _ = self._normalize_verts(verts)

        var_cmd = None
        # texture coords need transformation from variable value to [ST]
        if tcoords is not None:
            tcoords, var_cmd = self._build_st_coords(tcoords, verts.size // 3)

        self.session.log(f"Part '{self.cmd.name}' defined: {num_lines} lines.")
        command = self.cmd

        return command, verts, tcoords, var_cmd

    def point_rep(self):
        """
        This function processes the geometry arrays and returns values to represent point data

        Returns
        -------
        On failure, the method returns None for the first return value.  The returned tuple is:

        (part_command, vertices, sizes, colors, var_command)

        part_command: UPDATE_PART command object
        vertices: numpy array of per-node coordinates
        sizes: numpy array of per-node radii
        colors: numpy array of per-node rgb colors
        var_command: UPDATE_VARIABLE command object for the variable the colors correspond to, if any
        """
        if self.cmd is None:
            return None, None, None, None, None
        if self.cmd.render != self.cmd.NODES:
            # Early out.  Rendering type for this object is a surface rep, not a point rep
            return None, None, None, None, None
        verts = self.coords
        num_verts = verts.size // 3
        norm_scale = self._normalize_verts(verts)

        # Convert var values in self.tcoords to RGB colors
        # For now, look up RGB colors.  Planned USD enhancements should allow tex coords instead.
        colors = None
        var_cmd = None

        if self.tcoords.size and self.tcoords.size == num_verts:
            var_dsg_id = self.cmd.color_variableid
            var_cmd = self.session.variables[var_dsg_id]
            if len(var_cmd.levels) == 0:
                self.session.log(
                    f"Note: Node rep not created for part '{self.cmd.name}'.  It has var values, but a palette with 0 levels."
                )
                return None, None, None, None, None

            p_min = None
            p_max = None
            for lvl in var_cmd.levels:
                if (p_min is None) or (p_min > lvl.value):
                    p_min = lvl.value
                if (p_max is None) or (p_max < lvl.value):
                    p_max = lvl.value

            num_texels = int(len(var_cmd.texture) / 4)

            colors = numpy.ndarray((num_verts * 3,), dtype="float32")
            low_color = [c / 255.0 for c in var_cmd.texture[0:3]]
            high_color = [
                c / 255.0 for c in var_cmd.texture[4 * (num_texels - 1) : 4 * (num_texels - 1) + 3]
            ]
            if p_min == p_max:
                # Special case where palette min == palette max
                mid_color = var_cmd[4 * (num_texels // 2) : 4 * (num_texels // 2) + 3]
                for idx in range(num_verts):
                    val = self.tcoords[idx]
                    if val == p_min:
                        colors[idx * 3 : idx * 3 + 3] = mid_color
                    elif val < p_min:
                        colors[idx * 3 : idx * 3 + 3] = low_color
                    elif val > p_min:
                        colors[idx * 3 : idx * 3 + 3] = high_color
            else:
                for idx in range(num_verts):
                    val = self.tcoords[idx]
                    if val <= p_min:
                        colors[idx * 3 : idx * 3 + 3] = low_color
                    else:
                        pal_pos = (num_texels - 1) * (val - p_min) / (p_max - p_min)
                        pal_idx, pal_sub = divmod(pal_pos, 1)
                        pal_idx = int(pal_idx)

                        if pal_idx >= num_texels - 1:
                            colors[idx * 3 : idx * 3 + 3] = high_color
                        else:
                            col0 = var_cmd.texture[pal_idx * 4 : pal_idx * 4 + 3]
                            col1 = var_cmd.texture[4 + pal_idx * 4 : 4 + pal_idx * 4 + 3]
                            for ii in range(0, 3):
                                colors[idx * 3 + ii] = (
                                    col0[ii] * pal_sub + col1[ii] * (1.0 - pal_sub)
                                ) / 255.0
            self.session.log(f"Part '{self.cmd.name}' defined: {self.coords.size // 3} points.")

        node_sizes = None
        if self.node_sizes.size and self.node_sizes.size == num_verts:
            # Pass out the node sizes if there is a size-by variable
            node_size_default = self.cmd.node_size_default * norm_scale
            node_sizes = numpy.ndarray((num_verts,), dtype="float32")
            for ii in range(0, num_verts):
                node_sizes[ii] = self.node_sizes[ii] * node_size_default
        elif norm_scale != 1.0:
            # Pass out the node sizes if the model is normalized to fit in a unit cube
            node_size_default = self.cmd.node_size_default * norm_scale
            node_sizes = numpy.ndarray((num_verts,), dtype="float32")
            for ii in range(0, num_verts):
                node_sizes[ii] = node_size_default

        self.session.log(f"Part '{self.cmd.name}' defined: {self.coords.size // 3} points.")
        command = self.cmd

        return command, verts, node_sizes, colors, var_cmd


class UpdateHandler(object):
    """
    This class serves as the interface between a DSGSession and a hosting application.
    The DSGSession processes the general aspects of the gRPC pipeline and collects the
    various DSG objects into collections of: groups, variables, etc.  It also coalesces
    the individual array updates into a "Part" object which represents a single addressable
    mesh chunk.
    UpdateHandler methods are called as the various update happen, and it is called when
    a mesh chunk has been entirely defined.  In most scenarios, a subclass of UpdateHandler
    is passed to the DSGSession to handshake the mesh data to the application target.
    """

    def __init__(self) -> None:
        self._session: "DSGSession"

    @property
    def session(self) -> "DSGSession":
        """The session object this handler has been associated with"""
        return self._session

    @session.setter
    def session(self, session: "DSGSession") -> None:
        self._session = session

    def add_group(self, id: int, view: bool = False) -> None:
        """Called when a new group command has been added: self.session.groups[id]"""
        if view:
            self.session.log(f"Adding view: {self.session.groups[id]}")
        else:
            self.session.log(f"Adding group: {self.session.groups[id].name}")

    def add_variable(self, id: int) -> None:
        """Called when a new group command has been added: self.session.variables[id]"""
        self.session.log(f"Adding variable: {self.session.variables[id].name}")

    def finalize_part(self, part: Part) -> None:
        """Called when all the updates on a Part object have been completed.

        Note: this superclass method should be called after the subclass has processed
        the part geometry as the saved part command will be destroyed by this call.
        """
        if part.cmd:
            self.session.log(f"Part finalized: {part.cmd.name}")
        part.cmd = None

    def start_connection(self) -> None:
        """A new gRPC connection has been established:  self.session.grpc"""
        grpc = self.session.grpc
        self.session.log(f"gRPC connection established to: {grpc.host}:{grpc.port}")

    def end_connection(self) -> None:
        """The previous gRPC connection has been closed"""
        self.session.log("gRPC connection closed")

    def begin_update(self) -> None:
        """A new scene update is about to begin"""
        self.session.log("Begin update ------------------------")

    def end_update(self) -> None:
        """The scene update is complete"""
        self.session.log("End update ------------------------")

    def get_dsg_cmd_attribute(self, obj: Any, name: str, default: Any = None) -> Optional[str]:
        """Utility function to get an attribute from a DSG update object

        Note: UpdateVariable and UpdateGroup commands support generic attributes
        """
        return obj.attributes.get(name, default)

    def group_matrix(self, group: Any) -> Any:
        matrix = group.matrix4x4
        # The Case matrix is basically the camera transform.  In vrmode, we only want
        # the raw geometry, so use the identity matrix.
        if (
            self.get_dsg_cmd_attribute(group, "ENS_OBJ_TYPE") == "ENS_CASE"
        ) and self.session.vrmode:
            matrix = [
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
            ]
        return matrix


class DSGSession(object):
    def __init__(
        self,
        port: int = 12345,
        host: str = "127.0.0.1",
        security_code: str = "",
        verbose: int = 0,
        normalize_geometry: bool = False,
        vrmode: bool = False,
        time_scale: float = 1.0,
        handler: UpdateHandler = UpdateHandler(),
        session: Optional["Session"] = None,
        uds_path: Optional[str] = None,
        grpc_use_tcp_sockets: bool = False,
        grpc_allow_network_connections: bool = False,
        grpc_disable_tls: bool = False,
    ):
        """
        Manage a gRPC connection and link it to an UpdateHandler instance

        This class makes a DSG gRPC connection via the specified port and host (leveraging
        the passed security code).  As DSG protobuffers arrive, they are merged into Part
        object instances and the UpdateHandler is invoked to further process them.

        Parameters
        ----------
        port : int
            The port number the EnSight gRPC service is running on.
            The default is ``12345``.
        host : str
            Name of the host that the EnSight gRPC service is running on.
            The default is ``"127.0.0.1"``, which is the localhost.
        security_code : str
            Shared security code for validating the gRPC communication.
            The default is ``""``.
        verbose : int
            The verbosity level.  If set to 1 or higher the class will call logging.info
            for log output.  The default is ``0``.
        normalize_geometry : bool
            If True, the scene coordinates will be remapped into the volume [-1,-1,-1] - [1,1,1]
            The default is not to remap coordinates.
        vrmode : bool
            If True, do not include the EnSight camera in the generated view group. The default
            is to include the EnSight view in the scene transformations.
        time_scale : float
            All DSG protobuffers time values will be multiplied by this factor after
            being received.  The default is ``1.0``.
        handler : UpdateHandler
            This is an UpdateHandler subclass that is called back when the state of
            a scene transfer changes.  For example, methods are called when the
            transfer begins or ends and when a Part (mesh block) is ready for processing.
        uds_path: string
            The unix domain socket path if required for the gRPC connection
        grpc_use_tcp_sockets: bool, optional
            If using gRPC, and if True, then allow TCP Socket based connections
            instead of only local connections.
        grpc_allow_network_connections: bool, optional
            If using gRPC and using TCP Socket based connections, listen on all networks.
        grpc_disable_tls: bool, optional
            If using gRPC and using TCP Socket based connections, disable TLS.
        """
        super().__init__()
        if uds_path:
            self._grpc = ensight_grpc.EnSightGRPC(
                grpc_uds_pathname=uds_path,
                secret_key=security_code,
                grpc_use_tcp_sockets=grpc_use_tcp_sockets,
                grpc_allow_network_connections=grpc_allow_network_connections,
                grpc_disable_tls=grpc_disable_tls,
            )
        else:
            self._grpc = ensight_grpc.EnSightGRPC(
                port=port,
                host=host,
                secret_key=security_code,
                grpc_use_tcp_sockets=grpc_use_tcp_sockets,
                grpc_allow_network_connections=grpc_allow_network_connections,
                grpc_disable_tls=grpc_disable_tls,
            )
        self._session = session
        if self._session:
            self._session.set_dsg_session(self)
        self._callback_handler = handler
        self._verbose = verbose
        self._thread: Optional[threading.Thread] = None
        self._message_queue: queue.Queue = queue.Queue()  # Messages coming from EnSight
        self._dsg_queue: Optional[queue.SimpleQueue] = None  # Outgoing messages to EnSight
        self._shutdown = False
        self._dsg = None
        # Prevent the protobuffer queue from growing w/o limits.  The payload chunking is
        # around 4MB, so 200 buffers would be a bit less than 1GB.
        self._max_dsg_queue_size = int(os.environ.get("ANSYS_OV_SERVER_MAX_GRPC_QUEUE_SIZE", "200"))
        self._normalize_geometry = normalize_geometry
        self._vrmode = vrmode
        self._time_scale = time_scale
        self._time_limits = [
            sys.float_info.max,
            -sys.float_info.max,
        ]  # Min/max across all time steps
        self._mesh_block_count = 0
        self._variables: Dict[int, Any] = dict()
        self._groups: Dict[int, Any] = dict()
        self._part: Part = Part(self)
        self._scene_bounds: Optional[List] = None
        self._cur_timeline: List = [0.0, 0.0]  # Start/End time for current update
        self._callback_handler.session = self
        # log any status changes to this file.  external apps will be monitoring
        self._status_file = os.environ.get("ANSYS_OV_SERVER_STATUS_FILENAME", "")
        self._status = dict(status="idle", start_time=0.0, processed_buffers=0, total_buffers=0)
        self._pyensight_grpc_coming = False

    @property
    def scene_bounds(self) -> Optional[List]:
        return self._scene_bounds

    @property
    def mesh_block_count(self) -> int:
        return self._mesh_block_count

    @property
    def max_dsg_queue_size(self) -> int:
        return self._max_dsg_queue_size

    @max_dsg_queue_size.setter
    def max_dsg_queue_size(self, value: int) -> None:
        self._max_dsg_queue_size = value

    @property
    def vrmode(self) -> bool:
        return self._vrmode

    @vrmode.setter
    def vrmode(self, value: bool) -> None:
        self._vrmode = value

    @property
    def normalize_geometry(self) -> bool:
        return self._normalize_geometry

    @normalize_geometry.setter
    def normalize_geometry(self, value: bool) -> None:
        self._normalize_geometry = value

    @property
    def variables(self) -> dict:
        return self._variables

    @property
    def groups(self) -> dict:
        return self._groups

    @property
    def part(self) -> Part:
        return self._part

    @property
    def time_limits(self) -> List:
        return self._time_limits

    @property
    def cur_timeline(self) -> List:
        return self._cur_timeline

    @cur_timeline.setter
    def cur_timeline(self, timeline: List) -> None:
        self._cur_timeline = timeline
        self._time_limits[0] = min(self._time_limits[0], self._cur_timeline[0])
        self._time_limits[1] = max(self._time_limits[1], self._cur_timeline[1])

    @property
    def grpc(self) -> ensight_grpc.EnSightGRPC:
        return self._grpc

    def log(self, s: str, level: int = 0) -> None:
        """Log a string to the logging system

        If the message level is less than the current verbosity,
        emit the message.
        """
        if level < self._verbose:
            logging.info(s)

    @staticmethod
    def warn(s: str) -> None:
        """Issue a warning to the logging system

        The logging message is mapped to "warn" and cannot be blocked via verbosity
        checks.
        """
        logging.warning(s)

    @staticmethod
    def error(s: str) -> None:
        """Issue an error to the logging system

        The logging message is mapped to "error" and cannot be blocked via verbosity
        checks.
        """
        logging.error(s)

    def start(self) -> int:
        """Start a gRPC connection to an EnSight instance

        Make a gRPC connection and start a DSG stream handler.

        Returns
        -------
            0 on success, -1 on an error.
        """
        # Start by setting up and verifying the connection
        self._grpc.connect()
        if not self._grpc.is_connected():
            self.log(f"Unable to establish gRPC connection to: {self._grpc.host}:{self._grpc.port}")
            return -1
        # Streaming API requires an iterator, so we make one from a queue
        # it also returns an iterator.  self._dsg_queue is the input stream interface
        # self._dsg is the returned stream iterator.
        if self._dsg is not None:
            return 0
        self._dsg_queue = queue.SimpleQueue()
        self._dsg = self._grpc.dynamic_scene_graph_stream(
            iter(self._dsg_queue.get, None)  # type:ignore
        )
        self._thread = threading.Thread(target=self._poll_messages)
        if self._thread is not None:
            self._thread.start()
        self._callback_handler.start_connection()
        return 0

    def end(self):
        """Stop a gRPC connection to the EnSight instance"""
        self._callback_handler.end_connection()
        self._grpc.shutdown()
        self._shutdown = True
        self._thread.join()
        self._grpc.shutdown()
        self._dsg = None
        self._thread = None
        self._dsg_queue = None

    def is_shutdown(self):
        """Check the service shutdown request status"""
        return self._shutdown

    def _update_status_file(self, timed: bool = False):
        """
        Update the status file contents. The status file will contain the
        following json object, stored as: self._status

        {
        'status' : "working|idle",
        'start_time' : timestamp_of_update_begin,
        'processed_buffers' : number_of_protobuffers_processed,
        'total_buffers' : number_of_protobuffers_total,
        }

        Parameters
        ----------
        timed : bool, optional:
            if True, only update every second.

        """
        if self._status_file:
            current_time = time.time()
            if timed:
                last_time = self._status.get("last_time", 0.0)
                if current_time - last_time < 1.0:  # type: ignore
                    return
            self._status["last_time"] = current_time
            try:
                message = json.dumps(self._status)
                with open(self._status_file, "w") as status_file:
                    status_file.write(message)
            except IOError:
                pass  # Note failure is expected here in some cases

    def request_an_update(self, animation: bool = False, allow_spontaneous: bool = True) -> None:
        """Start a DSG update
        Send a command to the DSG protocol to "init" an update.

        Parameters
        ----------
        animation:
            if True, export all EnSight timesteps.
        allow_spontaneous:
            if True, allow EnSight to trigger async updates.
        """
        # Send an INIT command to trigger a stream of update packets
        cmd = dynamic_scene_graph_pb2.SceneClientCommand()
        cmd.command_type = dynamic_scene_graph_pb2.SceneClientCommand.INIT
        # Allow EnSight push commands, but full scene only for now...
        cmd.init.allow_spontaneous = allow_spontaneous
        cmd.init.include_temporal_geometry = animation
        cmd.init.allow_incremental_updates = False
        cmd.init.maximum_chunk_size = 1024 * 1024
        self._dsg_queue.put(cmd)  # type:ignore

    def _is_queue_full(self):
        if not self.max_dsg_queue_size:
            return False
        if self._pyensight_grpc_coming:
            return False
        return self._message_queue.qsize() >= self.max_dsg_queue_size

    def _poll_messages(self) -> None:
        """Core interface to grab DSG events from gRPC and queue them for processing

        This is run by a thread that is monitoring the dsg RPC call for update messages
        it places them in _message_queue as it finds them.  They are picked up by the
        main thread via get_next_message()
        """
        while not self._shutdown:
            try:
                self._message_queue.put(next(self._dsg))  # type:ignore
                # if the queue is getting too deep, wait a bit to avoid holding too
                # many messages (filling up memory)
                if self.max_dsg_queue_size:
                    while self._is_queue_full():
                        time.sleep(0.001)
            except Exception:
                self._shutdown = True
                self.log("DSG connection broken, calling exit")
                sys.exit(0)

    def _get_next_message(self, wait: bool = True) -> Any:
        """Get the next queued up protobuffer message

        Called by the main thread to get any messages that were pulled in from the
        dsg stream and placed here by _poll_messages()
        """
        try:
            return self._message_queue.get(block=wait)
        except queue.Empty:
            return None

    def _reset(self):
        self._variables = {}
        self._groups = {}
        self._part = Part(self)
        self._scene_bounds = None
        self._mesh_block_count = 0  # reset when a new group shows up

    def handle_one_update(self) -> None:
        """Monitor the DSG stream and handle a single update operation

        Wait until we get the scene update begin message.  From there, reset the current
        scene buckets and then parse all the incoming commands until we get the scene
        update end command.   At which point, save the generated stage (started in the
        view command handler).  Note: Parts are handled with an available bucket at all times.
        When a new part update comes in or the scene update end happens, the part is "finished".
        """
        # An update starts with a UPDATE_SCENE_BEGIN command
        cmd = self._get_next_message()
        while (cmd is not None) and (
            cmd.command_type != dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_SCENE_BEGIN
        ):
            # Look for a begin command
            cmd = self._get_next_message()

        # Start anew
        self._reset()
        self._callback_handler.begin_update()

        # Update our status
        self._status = dict(
            status="working", start_time=time.time(), processed_buffers=1, total_buffers=1
        )
        self._update_status_file()

        # handle the various commands until UPDATE_SCENE_END
        cmd = self._get_next_message()
        while (cmd is not None) and (
            cmd.command_type != dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_SCENE_END
        ):
            self._handle_update_command(cmd)
            self._status["processed_buffers"] += 1  # type: ignore
            self._status["total_buffers"] = self._status["processed_buffers"] + self._message_queue.qsize()  # type: ignore
            self._update_status_file(timed=True)
            cmd = self._get_next_message()

        # Flush the last part
        self._finish_part()

        self._callback_handler.end_update()

        # Update our status
        self._status = dict(status="idle", start_time=0.0, processed_buffers=0, total_buffers=0)
        self._update_status_file()

    def _handle_update_command(self, cmd: dynamic_scene_graph_pb2.SceneUpdateCommand) -> None:
        """Dispatch out a scene update command to the proper handler

        Given a command object, pull out the correct portion of the protobuffer union and
        pass it to the appropriate handler.

        Parameters
        ----------
        cmd:
            The command to be dispatched.
        """
        name = "Unknown"
        if cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.DELETE_ID:
            name = "Delete IDs"
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_PART:
            name = "Part update"
            tmp = cmd.update_part
            self._handle_part(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_GROUP:
            name = "Group update"
            tmp = cmd.update_group
            self._handle_group(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_GEOM:
            name = "Geom update"
            tmp = cmd.update_geom
            self._part.update_geom(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_VARIABLE:
            name = "Variable update"
            tmp = cmd.update_variable
            self._handle_variable(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_VIEW:
            name = "View update"
            tmp = cmd.update_view
            self._handle_view(tmp)
        elif cmd.command_type == dynamic_scene_graph_pb2.SceneUpdateCommand.UPDATE_TEXTURE:
            name = "Texture update"
        self.log(f"{name} --------------------------")

    def _finish_part(self) -> None:
        """Complete the current part

        There is always a part being modified.  This method completes the current part, committing
        it to the handler.
        """
        try:
            self._callback_handler.finalize_part(self.part)
        except Exception as e:
            import traceback

            self.warn(f"Error encountered while finalizing part geometry: {str(e)}")
            traceback_str = "".join(traceback.format_tb(e.__traceback__))
            logging.debug(f"Traceback: {traceback_str}")
        self._mesh_block_count += 1

    def _handle_part(self, part_cmd: Any) -> None:
        """Handle a DSG UPDATE_PART command

        Finish the current part and set up the next part.

        Parameters
        ----------
        part:
            The command coming from the EnSight stream.
        """
        self._finish_part()
        self._part.reset(part_cmd)

    def find_group_pb(self, group_id: int) -> Any:
        """Return the group command protobuffer for a specific group id.

        Parameters
        ----------
        group_id: int
            The group DSG protocol entity id.

        Returns
        -------
        any
            The group command protobuffer or None.
        """
        return self._groups.get(group_id, None)

    def _handle_group(self, group: Any) -> None:
        """Handle a DSG UPDATE_GROUP command

        Parameters
        ----------
        group:
            The command coming from the EnSight stream.
        """
        # reset current mesh (part) count for unique "part" naming in USD
        self._mesh_block_count = 0

        # record the scene bounds in case they are needed later
        self._groups[group.id] = group
        bounds = group.attributes.get("ENS_SCENE_BOUNDS", None)
        if bounds:
            minmax = list()
            for v in bounds.split(","):
                try:
                    minmax.append(float(v))
                except ValueError:
                    pass
            if len(minmax) == 6:
                self._scene_bounds = minmax
        # callback
        self._callback_handler.add_group(group.id)

    def _handle_variable(self, var: Any) -> None:
        """Handle a DSG UPDATE_VARIABLE command

        Save off the EnSight variable DSG command object.

        Parameters
        ----------
        var:
            The command coming from the EnSight stream.
        """
        self._variables[var.id] = var
        self._callback_handler.add_variable(var.id)

    def _handle_view(self, view: Any) -> None:
        """Handle a DSG UPDATE_VIEW command

        Parameters
        ----------
        view:
            The command coming from the EnSight stream.
        """
        self._finish_part()
        self._scene_bounds = None
        self._groups[view.id] = view
        if len(view.timeline) == 2:
            view.timeline[0] *= self._time_scale
            view.timeline[1] *= self._time_scale
            self.cur_timeline = [view.timeline[0], view.timeline[1]]
        self._callback_handler.add_group(view.id, view=True)
